"use strict";

exports.__esModule = true;
exports.default = generateSpec;

var _path2 = _interopRequireDefault(require("path"));

var _lodash = _interopRequireWildcard(require("lodash"));

var _fs = _interopRequireDefault(require("fs"));

var _uuid = _interopRequireDefault(require("uuid"));

function _getRequireWildcardCache() { if (typeof WeakMap !== "function") return null; var cache = new WeakMap(); _getRequireWildcardCache = function _getRequireWildcardCache() { return cache; }; return cache; }

function _interopRequireWildcard(obj) { if (obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function asyncGeneratorStep(gen, resolve, reject, _next, _throw, key, arg) { try { var info = gen[key](arg); var value = info.value; } catch (error) { reject(error); return; } if (info.done) { resolve(value); } else { Promise.resolve(value).then(_next, _throw); } }

function _asyncToGenerator(fn) { return function () { var self = this, args = arguments; return new Promise(function (resolve, reject) { var gen = fn.apply(self, args); function _next(value) { asyncGeneratorStep(gen, resolve, reject, _next, _throw, "next", value); } function _throw(err) { asyncGeneratorStep(gen, resolve, reject, _next, _throw, "throw", err); } _next(undefined); }); }; }

function _extends() { _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; }; return _extends.apply(this, arguments); }

//let paths = path.join(spice.root_path, "routes");
var tags = [];

var Comments = require("parse-comments");

var comments = new Comments();

function getPackageInfo() {
  var pac = require(_path2.default.join(spice.parent_app_dir, "/package.json"));

  return pac;
}

function createPath(path) {
  return _lodash.default.join(_lodash.default.map(_lodash.default.split(path, "/"), seg => {
    if (_lodash.default.startsWith(seg, ":")) {
      return "{" + _lodash.default.trim(seg, ":") + "}";
    }

    return seg;
  }), "/");
}

function getRoutes(models) {
  var paths = _path2.default.join(spice.root_path, "routes");

  var routes = [];

  require("fs").readdirSync(paths).forEach(function (file) {
    if (file != "spice.js") {
      var router = require(_path2.default.join(paths, file));

      var router_string = _fs.default.readFileSync(_path2.default.join(paths, file)).toString();

      var parsed_comments = comments.parse(router_string);

      var comment_processed = _lodash.default.map(parsed_comments, comment => {
        if (comment.code.value.split(".")[0] == "router") {
          var code_array = comment.code.value.split("\"");

          if (code_array.length == 1) {
            code_array = comment.code.value.split("'");
          }

          if (code_array[0] != undefined) {
            var router_method = code_array[0].split("(");
            var router_method_array = router_method[0].split(".");
            comment.path = createPath(code_array[1]);
            comment.original_path = code_array[1];
            comment.method = router_method_array[1];
            comment.path_method = comment.original_path + "-" + comment.method;
            return comment;
          }
        }
      }).filter(item => item != undefined);

      tags.push({
        name: file.split(".")[0],
        description: parsed_comments[0] != undefined ? parsed_comments[0].description : ""
      });
      var defaultModel = findDefaultModel(parsed_comments[0]);

      _lodash.default.each(router.stack, layer => {
        if (layer.methods.length > 0) {
          var _path = {
            path: createPath(layer.path),
            original_path: layer.path,
            method: _lodash.default.lowerCase(_lodash.default.last(layer.methods)),
            params: _lodash.default.map(layer.paramNames, param => {
              return {
                name: param.name,
                type: "string",
                required: !param.optional,
                in: "path"
              };
            }).filter(obj => obj != undefined),
            tags: [file.split(".")[0]],
            produces: ["application/json"]
          };
          _path.responses = {
            200: {
              description: "Successful Operation"
            },
            401: {
              description: "Authentication Error"
            },
            400: {
              description: "Bad Request"
            },
            404: {
              description: "Not found"
            }
          };

          var comment = _lodash.default.find(comment_processed, comment => {
            function matches() {
              var match = false;
              match = _lodash.default.endsWith(layer.path, comment.original_path);
              return match;
            }

            return comment.method == _lodash.default.lowerCase(_lodash.default.last(layer.methods)) && matches();
          });

          if (comment != undefined) {
            _path.description = comment.description;
            _path.operationId = _uuid.default.v4();
            _path.parameters = [...preparParameters(comment, undefined, models, router), ..._path.params];
            var doc_return = filterTags(comment, "returns");

            if (doc_return.length > 0) {
              if (doc_return[0].type == undefined) {
                _path.responses["200"]["schema"] = {
                  $ref: "#/definitions/" + doc_return[0].description
                };
              } else if (doc_return[0].type.name == "array") {
                _path.responses["200"]["schema"] = {
                  type: doc_return[0].type.name,
                  items: {
                    $ref: "#/definitions/" + doc_return[0].description
                  }
                };
              } else if (doc_return[0].type.name == "object") {
                _path.responses["200"]["schema"] = {
                  type: doc_return[0].type.name,
                  additionalProperties: {
                    $ref: "#/definitions/" + doc_return[0].description
                  }
                };
              }
            }
          }

          routes.push(_path);
        }
      }).filter(item => item != undefined);
    }
  });

  return routes;
}

function preparParameters(comment, defaultModel, models) {
  if (comment.method == "post" || comment.method == "put") {
    var param_list = filterTags(comment, "param");
    var route_model_params = [];

    var params_from_comments = _lodash.default.map(param_list, tag => {
      var param;

      if (tag.name != "") {
        param = {
          name: tag.name,
          type: "string",
          paramType: "form"
        };
      }

      if (tag.description != undefined && IsJsonString(tag.description)) {
        var parsed_JSON = JSON.parse(tag.description);
        param = _extends({}, param, parsed_JSON);
      }

      if (tag.name == "" && tag.type != undefined) {
        route_model_params.push(tag.type.name);
      }

      return param;
    }).filter(obj => obj != undefined);

    var params_from_model = [];

    if (defaultModel && param_list.length == 0) {
      params_from_model = _lodash.default.union(params_from_model, modelToParameter(defaultModel));
    } else if (route_model_params.length > 0) {
      _lodash.default.each(route_model_params, Model => {
        params_from_model = _lodash.default.union(params_from_model, modelToParameter(Model));
      });
    }

    function modelToParameter(model) {
      return _lodash.default.map(models[model], (value, key) => {
        return {
          name: key,
          type: _lodash.default.includes(["object", "array"], value.type) ? "string" : value.type,
          description: value.description ? value.description : "",
          required: value.required ? value.required : false,
          paramType: "form",
          in: "formData"
        };
      }).filter(obj => obj != undefined);
    }
    /* console.log("Props::::::: ", [
      ...params_from_comments,
      ...params_from_model,
    ]); */


    return [...params_from_comments, ...params_from_model];
  } else {
    return [];
  }
}

function IsJsonString(str) {
  try {
    JSON.parse(str);
  } catch (e) {
    return false;
  }

  return true;
}

function findDefaultModel(comment) {
  if (comment == undefined) {
    return undefined;
  } else {
    var id_found = _lodash.default.first(filterTags(comment, "default_model"));

    return id_found == undefined ? id_found : id_found.description;
  }
}

function findId(comment) {
  var id_found = _lodash.default.first(filterTags(comment, "id"));

  return id_found == undefined ? id_found : id_found.description;
}

function filterTags(comment, title) {
  return _lodash.default.filter(findTags(comment), tag => {
    return tag.title == title;
  });
}

function findTags(comment) {
  return comment.tags;
}

function getModels() {
  return _getModels.apply(this, arguments);
}

function _getModels() {
  _getModels = _asyncToGenerator(function* () {
    var paths = _path2.default.join(spice.root_path, "models");

    var models = {};

    require("fs").readdirSync(paths).forEach(function (file) {
      var model = require(_path2.default.join(paths, file));

      models[file.split(".")[0]] = model.props;
    });

    return models;
  });
  return _getModels.apply(this, arguments);
}

function getRequired(obj) {
  return _lodash.default.map(obj, (value, key) => {
    if (value.required == true) {
      return key;
    }
  }).filter(item => item != undefined);
}

function getProperties(obj) {
  return _lodash.default.mapValues(obj, (_ref) => {
    var {
      type,
      description,
      options,
      schema,
      map
    } = _ref;
    var to_return = {
      type: type,
      description: description,
      enum: options
    };

    if (to_return.type == "array") {
      to_return.items = {};
    }

    if (to_return.type == "date") {
      to_return.type = "string";
      to_return.format = "date";
    }

    if (map != undefined && map.type == "model") {
      if (map.reference != undefined) {
        var references = _lodash.default.isArray(map.reference) ? map.reference : [map.reference];
        var reference = "#/definitions/" + (_lodash.default.isString(map.reference) ? map.reference : references[0].name);

        if (to_return.type == "array") {
          to_return.items.$ref = reference;
        }

        if (to_return.type == "object") {
          to_return.$ref = reference;
        }

        if (to_return.type == "string") {
          to_return.type = "object";
          to_return.readOnly = true;
          to_return.$ref = reference;
        }
      }
    }

    return to_return;
  });
}

function generateDefinitions(items) {
  var models = _lodash.default.cloneDeep(items);

  _lodash.default.each(models, (value, key) => {
    _lodash.default.each(value, (val, key2) => {
      if (val.props != undefined) {
        //console.log(key, key2, val, value);
        models[key2] = val.props;
        value[key2].map = {
          type: "model",
          reference: key2
        };
      }
    });
  });

  var values = _lodash.default.mapValues(models, obj => {
    return {
      type: "object",
      required: getRequired(obj),
      properties: getProperties(obj)
    };
  });

  return values;
}

function generatePaths(routes) {
  var json = _lodash.default.groupBy(routes, "path");

  _lodash.default.forEach(json, function (value, key) {
    json[key] = _lodash.default.groupBy(json[key], "method");

    _lodash.default.forEach(json[key], function (inside_value, inside_key) {
      json[key][inside_key] = inside_value[0];
    });
  });

  return json;
}

function generateTags() {
  return _lodash.default.map(tags, tag => {
    return {
      name: tag.name,
      description: tag.description
    };
  });
}

function generateSpec() {
  return _generateSpec.apply(this, arguments);
}

function _generateSpec() {
  _generateSpec = _asyncToGenerator(function* () {
    var {
      description,
      version,
      name
    } = getPackageInfo();

    var models = _lodash.default.cloneDeep((yield getModels()));

    var routes = getRoutes(models);
    var spec = {
      swagger: "2.0",
      info: {
        description: description,
        version: version,
        title: name,
        contact: {
          name: spice.config.spice.developer.name || "Sonover Inc.",
          email: spice.config.spice.developer.email
        }
      },
      securityDefinitions: {
        Bearer: {
          type: "apiKey",
          name: "Authorization",
          in: "header"
        }
      },
      security: [{
        Bearer: []
      }],
      host: spice.config.spice.host + ":" + (spice.config.spice.external_port || spice.config.spice.port),
      port: spice.config.spice.external_port || spice.config.spice.port,
      basePath: "/",
      tags: generateTags(),
      paths: generatePaths(routes),
      definitions: generateDefinitions(models)
    };
    return spec;
  });
  return _generateSpec.apply(this, arguments);
}