"use strict";

exports.__esModule = true;
exports.default = exports.status = exports.addTask = exports.MapType = exports.SpiceCache = exports.MailFile = exports.Storage = exports.LocalStorage = exports.DebugStorage = exports.MailDebug = exports.DataType = exports.Mail = exports.RestHelper = exports.Serializer = exports.Crypt = exports.SpiceModel = exports.ResourceLifecycleTriggered = exports.SocketEventDriver = exports.SpiceEventDriver = exports.EventDebugger = exports.Event = void 0;

var _Event = _interopRequireDefault(require("./events/Event"));

exports.Event = _Event.default;

var _EventDebugger = _interopRequireDefault(require("./events/providers/EventDebugger"));

exports.EventDebugger = _EventDebugger.default;

var _SpiceEventDriver = _interopRequireDefault(require("./events/providers/SpiceEventDriver"));

exports.SpiceEventDriver = _SpiceEventDriver.default;

var _SocketEventDriver = _interopRequireDefault(require("./events/providers/SocketEventDriver"));

exports.SocketEventDriver = _SocketEventDriver.default;

var _ResourceLifecycleTriggered = _interopRequireDefault(require("./events/events/ResourceLifecycleTriggered"));

exports.ResourceLifecycleTriggered = _ResourceLifecycleTriggered.default;

var _SpiceModel = _interopRequireDefault(require("./models/SpiceModel"));

exports.SpiceModel = _SpiceModel.default;

var _Crypt = _interopRequireDefault(require("./utility/Crypt"));

exports.Crypt = _Crypt.default;

var _Serializer = _interopRequireDefault(require("./utility/Serializer"));

exports.Serializer = _Serializer.default;

var _RestHelper = _interopRequireDefault(require("./utility/RestHelper"));

exports.RestHelper = _RestHelper.default;

var _Status = _interopRequireDefault(require("./utility/Status"));

var _Mail = _interopRequireDefault(require("./mail/Mail"));

exports.Mail = _Mail.default;

var _DataType = _interopRequireDefault(require("./utility/DataType"));

exports.DataType = _DataType.default;

var _Debug = _interopRequireDefault(require("./mail/providers/Debug"));

exports.MailDebug = _Debug.default;

var _Debug2 = _interopRequireDefault(require("./storage/providers/Debug"));

exports.DebugStorage = _Debug2.default;

var _Local = _interopRequireDefault(require("./storage/providers/Local"));

exports.LocalStorage = _Local.default;

var _Storage = _interopRequireDefault(require("./storage/Storage"));

exports.Storage = _Storage.default;

var _File = _interopRequireDefault(require("./mail/providers/File"));

exports.MailFile = _File.default;

var _SpiceCache = _interopRequireDefault(require("./cache/providers/SpiceCache"));

exports.SpiceCache = _SpiceCache.default;

var _MapType = _interopRequireDefault(require("./utility/MapType"));

exports.MapType = _MapType.default;

var _tasks = _interopRequireDefault(require("./bootstrap/tasks"));

exports.addTask = _tasks.default;

var _koa2SwaggerUi = require("koa2-swagger-ui");

var _generator = _interopRequireDefault(require("./docs/generator"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function asyncGeneratorStep(gen, resolve, reject, _next, _throw, key, arg) { try { var info = gen[key](arg); var value = info.value; } catch (error) { reject(error); return; } if (info.done) { resolve(value); } else { Promise.resolve(value).then(_next, _throw); } }

function _asyncToGenerator(fn) { return function () { var self = this, args = arguments; return new Promise(function (resolve, reject) { var gen = fn.apply(self, args); function _next(value) { asyncGeneratorStep(gen, resolve, reject, _next, _throw, "next", value); } function _throw(err) { asyncGeneratorStep(gen, resolve, reject, _next, _throw, "throw", err); } _next(undefined); }); }; }

var path = require("path"); //import IO from "koa-socket-2";


var cors = require("kcors");

global.spice = {};

require("dotenv").config();

var Validate = require("koa-validation");

var koaBody = require("koa-body");

spice.parent_app_dir = path.resolve("./");
spice.root_path = path.resolve("./build/");
spice.module_root_path = path.resolve(__dirname);

var convert = require("koa-convert");

spice.classes = {};
/* spice.classes.Event = Event;
spice.classes.Serializer = Serializer;
spice.classes.RestHelper = RestHelper; */

spice.classes.validate = Validate;

/* spice.classes.Mail = Mail; */
spice.classes.mail = {
  providers: {}
};
var status = _Status.default;
exports.status = status;

class Spice {
  constructor() {}

  init(app) {
    return _asyncToGenerator(function* () {
      try {
        //const io = new IO();
        app.use(cors());

        require("koa-qs")(app, "extended");

        app.use(koaBody({
          formLimit: "2gb",
          jsonLimit: "1gb",
          textLimit: "1gb",
          multipart: true,
          formidable: {
            uploadDir: "./storage/uploads",
            maxFileSize: 1024 * 1024 * 1024 // 1GB

          }
        }));
        app.use(convert(Validate()));
        spice.app = app;
        spice.mofifiers = {};
        spice.config = yield require("./loaders/config")();

        spice.addModifier = function (resource, modifier) {
          spice.mofifiers[resource.toLowerCase()] = spice.mofifiers[resource] == undefined ? [modifier] : [...spice.mofifiers[resource], modifier];
        };

        spice.getModifiers = function (resource) {
          return spice.mofifiers[resource.toLowerCase()] || [];
        }; //io.attach(app);
        //spice.io = io; */

        /* io.on("message", (ctx, data) => {
          console.log("client sent data to message endpoint", data);
        }); */

        /* app._io.on("connection", (sock) => {
          console.log("Connection Up", sock);
        }); */


        app.use((0, _koa2SwaggerUi.koaSwagger)({
          hideTopbar: true,
          title: "Spice JS",
          routePrefix: "/docs",
          // host at /swagger instead of default /docs
          swaggerOptions: {
            spec: yield (0, _generator.default)(),
            deepLinking: true,
            displayRequestDuration: true,
            jsonEditor: true
          },
          exposeSpec: true
        }));
        yield require("./loaders").load();
        return spice;
      } catch (e) {
        console.log(e.stack);
      }
    })();
  }

}

exports.default = Spice;