"use strict";

exports.__esModule = true;
exports.default = void 0;

var _2 = require("..");

var _ResourceLifecycleTriggered = _interopRequireDefault(require("../events/events/ResourceLifecycleTriggered"));

var _Security = require("../utility/Security");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _extends() { _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; }; return _extends.apply(this, arguments); }

function asyncGeneratorStep(gen, resolve, reject, _next, _throw, key, arg) { try { var info = gen[key](arg); var value = info.value; } catch (error) { reject(error); return; } if (info.done) { resolve(value); } else { Promise.resolve(value).then(_next, _throw); } }

function _asyncToGenerator(fn) { return function () { var self = this, args = arguments; return new Promise(function (resolve, reject) { var gen = fn.apply(self, args); function _next(value) { asyncGeneratorStep(gen, resolve, reject, _next, _throw, "next", value); } function _throw(err) { asyncGeneratorStep(gen, resolve, reject, _next, _throw, "throw", err); } _next(undefined); }); }; }

var co = require("co");

var regeneratorRuntime = require("regenerator-runtime");

var SDate = require("sonover-date"),
    UUID = require("uuid"),
    _ = require("lodash"),
    _database = Symbol(),
    _ctx = Symbol(),
    _props = Symbol(),
    _args = Symbol(),
    _hooks = Symbol(),
    _disable_lifecycle_events = Symbol(),
    _external_modifier_loaded = Symbol(),
    _skip_cache = Symbol(),
    _serializers = Symbol(); //const _type = Symbol("type");


var that;

if (!Promise.allSettled) {
  Promise.allSettled = promises => Promise.all(promises.map((promise, i) => promise.then(value => ({
    status: "fulfilled",
    value
  })).catch(reason => ({
    status: "rejected",
    reason
  }))));
}

class SpiceModel {
  constructor(args) {
    if (args === void 0) {
      args = {};
    }

    try {
      var _args$args, _args2, _args2$args, _args3, _args3$args;

      var dbtype = spice.config.database.connections[args.connection].type || "couchbase";

      var Database = require("spice-" + dbtype);

      this.type = "";
      this.collection = args.connection;
      this[_args] = args.args;
      this[_external_modifier_loaded] = false;
      this[_disable_lifecycle_events] = ((_args$args = args.args) == null ? void 0 : _args$args.disable_lifecycle_events) || false;
      this[_ctx] = (_args2 = args) == null ? void 0 : (_args2$args = _args2.args) == null ? void 0 : _args2$args.ctx;
      this[_skip_cache] = ((_args3 = args) == null ? void 0 : (_args3$args = _args3.args) == null ? void 0 : _args3$args.skip_cache) || false;
      this[_hooks] = {
        create: {
          before: [],
          after: []
        },
        get: {
          before: [],
          after: []
        },
        list: {
          before: [],
          after: []
        },
        update: {
          before: [],
          after: []
        },
        delete: {
          before: [],
          after: []
        }
      };
      this[_serializers] = {
        read: {
          modifiers: [],
          cleaners: ["deleted", "type"]
        },
        write: {
          modifiers: [],
          cleaners: []
        }
      };
      this.deleted = false;
      this[_database] = new Database(args.connection || "default", {
        collection: args.collection,
        scope: args.scope
      });

      function removeDynamicProps(props) {
        var returnVal = {};

        for (var i in props) {
          if (!_.has(props[i], "source") || _.get(props[i], "source") == "static") {
            returnVal[i] = props[i];
          }
        }

        return returnVal;
      }

      function applySchemaOverrides(props, resource) {
        spice.schema_extenders = spice.schema_extenders || [];

        function applyExtenderModifier(prop, resource, extender) {
          if (extender.modifier) {
            return extender.modifier(prop, resource);
          }

          return prop;
        }

        function hasExtender(resource, prop, extender) {
          if (extender.selector(prop, resource)) {
            return true;
          }

          return false;
        }

        function resourceHasExtender(resource, extender) {
          return extender.resource.includes(resource) || extender.resource.includes("*");
        }

        for (var file_name in spice.schema_extenders) {
          var extender = spice.schema_extenders[file_name];

          if (resourceHasExtender(resource, extender)) {
            for (var j in props) {
              if (hasExtender(resource, props[j], extender)) {
                props[j] = applyExtenderModifier(props[j], resource, extender);
              }
            }

            if (extender == null ? void 0 : extender.creator) {
              props = _.merge(props, extender.creator(resource));
            }
          }
        }

        return props;
      } //console.log("Applying Overrides to :", this.collection,'||', args.collection)


      args.props = applySchemaOverrides(removeDynamicProps(args.props), args.collection);
      this[_props] = args.props;

      for (var i in args.props) {
        if (args.args != undefined) {
          if (args.args[i] != undefined) {
            switch (args.props[i].type) {
              case String:
              case "string":
                {
                  this[i] = args.args[i];
                  break;
                }

              case Number:
              case "number":
                {
                  this[i] = _.isNumber(args.args[i]) ? args.args[i] : Number(args.args[i]);
                  break;
                }

              case Boolean:
              case "boolean":
                {
                  this[i] = _.isBoolean(args.args[i]) ? args.args[i] : args.args[i] == "true" || args.args[i] == 1 || args.args[i] == "True";
                  break;
                }

              case Date:
              case "date":
                {
                  this[i] = args.args[i];
                  break;
                }

              case Array:
              case "array":
                {
                  this[i] = _.isArray(args.args[i]) ? args.args[i] : JSON.parse(args.args[i]);
                  break;
                }

              case Object:
              case "object":
                {
                  this[i] = _.isObject(args.args[i]) ? args.args[i] : JSON.parse(args.args[i]);
                  break;
                }

              default:
                {
                  this[i] = args.args[i];
                }
            }
          }
        }
      }

      that = this;
      this.createModifiersFromProps();
    } catch (e) {
      console.log(e.stack);
    } // }

  }

  get database() {
    return this[_database];
  }

  set database(value) {
    this[_database] = value;
  }

  get props() {
    return this[_props];
  }

  set props(value) {
    this[_props] = value;
  }

  is_date(item) {
    return item.length > 27 && item.indexOf("|") > -1;
  }

  propsToBeRemoved(data) {
    var _this = this;

    return _asyncToGenerator(function* () {
      if (!_.isArray(data)) {
        data = Array.of(data);
      }

      if (_this[_ctx]) {
        var _this$_ctx, _this$_ctx$state;

        if ((_this$_ctx = _this[_ctx]) == null ? void 0 : (_this$_ctx$state = _this$_ctx.state) == null ? void 0 : _this$_ctx$state.process_fields) {
          var returned = yield new _this[_ctx].state.process_fields().process(_this[_ctx], data, _this.type);
          return returned;
        } else {
          return data;
        }
      }

      return [];
    })();
  }

  createValidationString(_ref) {
    var {
      property_name,
      props,
      operation,
      complete
    } = _ref;
    var operationExempt = {
      PUT: ["required", "unique"]
    };
    var returnVal = [];

    for (var i in props) {
      if (!_.includes(operationExempt[operation] || [], i) || complete) {
        if (i == "required" && props[i] == true) {
          returnVal.push("required");
        }

        if (i == "email" && props[i] == true) {
          returnVal.push("email");
        }

        if (i == "json" && props[i] == true) {
          returnVal.push("json");
        }

        if (i == "ip" && props[i] == true) {
          returnVal.push("ip");
        }

        if (i == "url" && props[i] == true) {
          returnVal.push("url");
        }

        if (i == "date" && props[i] == true) {
          returnVal.push("date");
        }

        if (i == "requires") {
          returnVal.push("shouldExist:" + props[i]);
        }

        if (i == "unique" && props[i] == true) {
          returnVal.push("unique:" + this.constructor.name + "," + property_name);
        }

        if (i == "type") {
          switch (props[i]) {
            case String:
            case _2.DataType.STRING:
              {
                returnVal.push("string");
                break;
              }

            case Number:
            case _2.DataType.NUMBER:
              {
                //returnVal.push(`numeric`);
                break;
              }

            case Boolean:
            case _2.DataType.BOOLEAN:
              {
                returnVal.push("in:true,false,True,False,TRUE,FALSE,1,0");
                break;
              }

            case Date:
            case _2.DataType.DATE:
              {
                returnVal.push("date");
                break;
              }

            case Object:
            case _2.DataType.OBJECT:
              {
                //returnVal.push(`json`);
                break;
              }

            case Array:
            case _2.DataType.ARRAY:
              {
                //returnVal.push(`json`);
                break;
              }

            default:
              {
                returnVal.push("string");
                break;
              }
          }
        }
      }
    }

    return _.join(returnVal, "|");
  }

  createFilterString(property, props) {
    var returnVal = [];

    for (var i in props) {
      if (i == "lowercase" && props[i] == true) {
        returnVal.push("lowercase");
      }

      if (i == "trim" && props[i] == true) {
        returnVal.push("trim");
      }

      if (i == "uppercase" && props[i] == true) {
        returnVal.push("uppercase");
      }
    }

    return _.join(returnVal, "|");
  }

  getValidationRules(method, _temp) {
    if (method === void 0) {
      method = "POST";
    }

    var {
      complete = false,
      pick = [],
      omit = []
    } = _temp === void 0 ? {} : _temp;
    var rules = {};
    var filters = {};
    var working_properties = this.props;

    if (pick.length > 0) {
      working_properties = _.pick(this.props, pick);
    }

    if (omit.length > 0) {
      working_properties = _.omit(this.props, omit);
    }

    for (var i in working_properties) {
      var created_rules = this.createValidationString({
        property_name: i,
        props: this.props[i],
        operation: method,
        complete
      });

      if (created_rules.length > 0) {
        rules[i] = created_rules;
      }

      var created_filter = this.createFilterString(i, this.props[i]);

      if (created_filter.length > 0) {
        filters[i] = created_filter;
      }
    }

    return {
      rules,
      messages: {},
      filters: {
        before: filters
      }
    };
  }

  makeQueryFromFilter(filter) {
    var return_string = "";

    for (var key in filter) {
      var item_string = "";

      if (filter[key].length > 1) {
        item_string = "(";
      }

      var item_count = 0;

      for (var item of filter[key]) {
        if (item_count > 0) {
          item_string = item_string + " OR ";
        }

        if (this.is_date(item)) {
          var to_index = item.indexOf("|");
          var first = item.substring(0, to_index);
          var second = item.substring(to_index + 1, item.length);
          item_string = item_string + " " + key + " BETWEEN '" + first + "' AND '" + second + "'";
        } else {
          item_string = item_string + " " + key + " LIKE '" + item + "'";
        }

        item_count++;
      }

      if (filter[key].length > 1) {
        item_string = item_string + ")";
      }

      if (return_string != "") {
        return_string = return_string + " AND ";
      }

      return_string = return_string + item_string;
    }

    return return_string;
  }

  shouldUseCache(resource_type) {
    var _spice$config$cache;

    // If '_skip_cache' property of this object is true, then we shouldn't cache.
    if (this[_skip_cache] == true) {
      return false;
    } // If the system configuration for spice has a cache status set to "disable",
    // then we shouldn't cache, so return false.


    if (((_spice$config$cache = spice.config.cache) == null ? void 0 : _spice$config$cache.status) == "disabled") {
      return false;
    } // If 'spice.cache[resource_type]' is not undefined,
    // it implies that this resource type is already in the cache or is cacheable.


    return spice.cache[resource_type] != undefined;
  }

  getCacheConfig(resource_type) {
    return spice.cache[resource_type] || {};
  }

  getCacheProviderObject(resource_type) {
    return spice.cache_providers[this.getCacheConfig(resource_type).driver || this.getCacheConfig(resource_type).provider || spice.config.cache.default_driver];
  }

  exists(item_type, key) {
    var _this2 = this;

    return _asyncToGenerator(function* () {
      var obj = _this2.getCacheProviderObject(item_type);

      if (obj) return yield obj.exists(key);
      return false;
    })();
  }

  shouldForceRefresh(response) {
    var _this3 = this;

    return _asyncToGenerator(function* () {
      var obj = _this3.getCacheProviderObject();

      var monitor_record = yield obj.get("monitor::" + _this3.type);

      if (monitor_record == undefined) {
        return false;
      }

      if (monitor_record.time > (response == null ? void 0 : response.time)) {
        return true;
      }

      return false;
    })();
  }

  setMonitor() {
    var current_time = new Date().getTime();
    var obj = this.getCacheProviderObject();
    var key = "monitor::" + this.type;
    var value = {
      time: current_time
    };
    obj.set(key, value, {
      ttl: 0
    });
  }

  get(args) {
    var _this4 = this;

    return _asyncToGenerator(function* () {
      try {
        if (!args) {
          args = {};
        }

        if (_.isString(args.id)) {
          if (args.skip_hooks != true) {
            yield _this4.run_hook(args, "get", "before");
          }

          var key = "get::" + _this4.type + "::" + args.id;
          var results = {};

          if (_this4.shouldUseCache(_this4.type)) {
            var cached_results = yield _this4.getCacheProviderObject(_this4.type).get(key);
            results = cached_results == null ? void 0 : cached_results.value;

            if ((cached_results == null ? void 0 : cached_results.value) == undefined || (yield _this4.shouldForceRefresh(cached_results)) || _this4.type == "workflow") {
              results = yield _this4.database.get(args.id);
              yield _this4.getCacheProviderObject(_this4.type).set(key, {
                value: results,
                time: new Date().getTime()
              }, _this4.getCacheConfig(_this4.type));
            }
          } else {
            results = yield _this4.database.get(args.id);
          }

          if (results.type != undefined) {
            if (results.type != _this4.type) throw new Error(_this4.type + " does not exist type");
          }

          if (results._type != undefined) {
            if (results._type != _this4.type) throw new Error(_this4.type + " does not exist _type");
          }

          if (results.deleted == undefined || results.deleted == false) {
            if (args.skip_read_serialize != true && args.skip_serialize != true) {
              results = yield _this4.do_serialize(results, "read", {}, args, (yield _this4.propsToBeRemoved(results)));
            }

            if (args.skip_hooks != true) {
              yield _this4.run_hook(results, "get", "after");
            }

            return results;
          } else {
            throw new Error(_this4.type + " does not exist");
          }
        }
      } catch (e) {
        console.log(e.message, e);
        throw e;
      }
    })();
  }

  query(query, scope) {
    var _this5 = this;

    return _asyncToGenerator(function* () {
      try {
        var results = yield _this5.database.query(query, scope);
        return results;
      } catch (error) {
        throw error;
      }
    })();
  }

  getMulti(args) {
    var _this6 = this;

    return _asyncToGenerator(function* () {
      try {
        if (!args) {
          args = {};
        }

        if (args.skip_hooks != true) {
          yield _this6.run_hook(_this6, "list", "before");
        }

        _.remove(args.ids, o => o == undefined);

        var key = "multi-get::" + _this6.type + "::" + _.join(args.ids, "|");

        var results = [];

        if (args.ids.length > 0) {
          if (_this6.shouldUseCache(_this6.type)) {
            var cached_results = yield _this6.getCacheProviderObject(_this6.type).get(key);
            results = cached_results == null ? void 0 : cached_results.value;

            if ((cached_results == null ? void 0 : cached_results.value) == undefined || (yield _this6.shouldForceRefresh(cached_results))) {
              results = yield _this6.database.multi_get(args.ids, true);

              _this6.getCacheProviderObject(_this6.type).set(key, {
                value: results,
                time: new Date().getTime()
              }, _this6.getCacheConfig(_this6.type));
            }
          } else {
            results = yield _this6.database.multi_get(args.ids, true);
          }
        }

        _.remove(results, o => o.type != _this6.type);

        if (args.skip_read_serialize != true && args.skip_serialize != true) {
          results = yield _this6.do_serialize(results, "read", {}, args, (yield _this6.propsToBeRemoved(results)));
        }

        if (args.skip_hooks != true) {
          yield _this6.run_hook(results, "list", "after", args.context);
        }

        return results;
      } catch (e) {
        console.log(e.stack);
        throw e;
      }
    })();
  }

  exist(data) {
    var _this7 = this;

    return _asyncToGenerator(function* () {
      try {
        if (_.isString(data)) {
          var result = yield _this7.database.get(data);
          if (result.type) if (result.type != _this7.type) {
            return false;
          }
          if (result._type) if (result._type != _this7.type) {
            return false;
          }
        } else {
          if (data.type) if (data.type != _this7.type) {
            return false;
          }
          if (data._type) if (data._type != _this7.type) {
            return false;
          }
        }

        return true;
      } catch (e) {
        throw e;
      }
    })();
  }

  update(args) {
    var _this8 = this;

    return _asyncToGenerator(function* () {
      try {
        _this8.updated_at = new SDate().now();
        var results = yield _this8.database.get(args.id);
        var item_exist = yield _this8.exist(results);

        if (!item_exist) {
          throw new Error(_this8.type + " does not exist. in update");
        }

        delete results["id"];

        _.defaults(_this8, results);

        var cover_obj = {
          old: results,
          new: _this8,
          id: args.id
        };
        var form;

        if (args.skip_write_serialize != true && args.skip_serialize != true) {
          cover_obj.new = yield _this8.do_serialize(_this8, "write", cover_obj.new, args);
        }

        if (args.skip_hooks != true) {
          yield _this8.run_hook(cover_obj, "update", "before", results);
        }

        var db_data = cover_obj.new || _this8;
        yield _this8.database.update(args.id, db_data);

        _this8.setMonitor();

        if (args.skip_read_serialize != true && args.skip_serialize != true) {
          cover_obj.new = yield _this8.do_serialize(cover_obj.new, "read", {}, args, (yield _this8.propsToBeRemoved(cover_obj.new)));
        }

        if (args.skip_hooks != true) {
          yield _this8.run_hook(_extends({}, _this8, {
            id: args.id
          }), "update", "after", results);
        }

        _this8.id = args.id;
        return _extends({}, cover_obj.new, {
          id: args.id
        });
      } catch (e) {
        console.log("Error on update", e, e.stack);
        throw e;
      }
    })();
  }

  create(args) {
    var _this9 = this;

    return _asyncToGenerator(function* () {
      if (args === void 0) {
        args = {};
      }

      try {
        var form;
        _this9.created_at = new SDate().now();
        args = _.defaults(args, {
          id_prefix: _this9.type
        });

        if (args.body) {
          form = _.defaults({}, args.body);
          form.created_at = _this9.created_at;
          form.updated_at = _this9.created_at;
          delete form["bucket"];
        }

        var workingForm = form || _this9;
        _this9.updated_at = new SDate().now();
        var id = args.id_prefix + "-" + UUID.v4();

        if (args && args.id) {
          id = args.id;
        }

        if (args.skip_write_serialize != true && args.skip_serialize != true) {
          workingForm = yield _this9.do_serialize(workingForm, "write", {}, args);
        }

        if (args.skip_hooks != true) {
          yield _this9.run_hook(workingForm, "create", "before");
        }

        var results = yield _this9.database.insert(id, workingForm, args.expiry);

        _this9.setMonitor();

        if (args.skip_read_serialize != true && args.skip_serialize != true) {
          results = yield _this9.do_serialize(results, "read", {}, args, (yield _this9.propsToBeRemoved(results)));
        }

        if (args.skip_hooks != true) {
          yield _this9.run_hook(_extends({}, results, {
            id
          }), "create", "after");
        }

        return _extends({}, results, {
          id
        });
      } catch (e) {
        console.log(e.stack);
        throw e;
      }
    })();
  }

  delete(args) {
    var _this10 = this;

    return _asyncToGenerator(function* () {
      var item_exist = yield _this10.exist(args.id);

      if (!item_exist) {
        throw new Error(_this10.type + " does not exist.");
      }

      var results = yield _this10.database.get(args.id);

      try {
        if (args.skip_hooks != true) {
          yield _this10.run_hook(args, "delete", "before");
        }

        var delete_response = {};

        if (args.hard) {
          delete_response = yield _this10.database.delete(args.id);

          _this10.setMonitor();
        } else {
          delete results["id"];
          results.deleted = true;
          delete_response = yield _this10.database.update(args.id, "");
        }

        if (args.skip_hooks != true) {
          yield _this10.run_hook(results, "delete", "after", results);
        }

        return {};
      } catch (e) {
        console.log(e.stack);
        throw e;
      }
    })();
  }

  toString() {
    JSON.stringify(this);
  }

  list(args) {
    var _this11 = this;

    return _asyncToGenerator(function* () {
      try {
        var _args4, _args5, _args6;

        if (!args) {
          args = {};
        }

        function prepColumns(columns) {
          if (columns && columns != "") {
            var columnList = columns.split(","); // Convert comma-separated string to array

            return _.join(_.compact(_.map(columnList, column => {
              if (column == "meta().id") {
                column = undefined;
              } else if (!column.startsWith("`") && !column.endsWith("`")) {
                "`" + column + "`";
              }

              if (column) {
                if (!column.includes(".") && !column.startsWith(that.type)) {
                  column = "`" + that.type + "`." + column;
                }
              }

              return column;
            })), ",");
          }

          return columns;
        }

        var query = "";
        args.columns = prepColumns(args.columns);

        function filterResultsByColumns(data, columns) {
          if (columns && columns != "") {
            columns = columns.replace(/`/g, "");
            columns = columns.replace(/meta\(\)\.id/g, "id");
            var columnList = columns.split(",").map(column => {
              //if colums has a . then remove the first part of the string
              return column.includes(".") ? _.last(column.split(".")) : column;
            }); // Convert comma-separated string to array

            columnList.push("_permissions"); // Always include _permissions

            columnList.push("_permissions_"); // Always include _permissions

            columnList.push("id"); // Always include _permissions

            return _.map(data, entry => _.pick(entry, columnList));
          }

          return data;
        }

        function extractNestings(string, localType) {
          var returnVal = [];
          var regex = /(\w+)\./g;
          var match;

          while (match = regex.exec(string)) {
            if (match[1] != localType) returnVal.push(match[1]);
          }

          return returnVal;
        }

        var nestings = [...extractNestings((_args4 = args) == null ? void 0 : _args4.query, _this11.type), ...extractNestings((_args5 = args) == null ? void 0 : _args5.columns, _this11.type), ...extractNestings((_args6 = args) == null ? void 0 : _args6.sort, _this11.type)];

        var mappedNestings = _.compact(_.map(_.uniq(nestings), nesting => {
          var prop = _this11.props[nesting];

          if (prop) {
            if (prop.map) {
              if (prop.map.type == _2.MapType.MODEL) {
                return {
                  alias: nesting,
                  reference: prop.map.reference.toLowerCase(),
                  value_field: prop.map.value_field
                };
              }
            }
          }
        }));

        function createJoinSection(nestings) {
          //create join section with nestings
          var joinSection = "";

          _.each(nestings, nesting => {
            joinSection += "LEFT JOIN `" + nesting.reference + "` AS `" + nesting.alias + "` ON KEYS `" + that.type + "`.`" + nesting.alias + "` ";
          });

          return joinSection;
        }

        var _join = createJoinSection(mappedNestings);

        args._join = _join;

        if (args.is_full_text && args.is_full_text === "true" || args.is_custom_query && args.is_custom_query === "true") {
          query = args.query;
        } else {
          if (args.filters) {
            query = _this11.makeQueryFromFilter(args.filters);
          } else {
            if (args.query) {
              query = args.query + (" AND (`" + _this11.type + "`.deleted = false OR `" + _this11.type + "`.deleted IS MISSING) ");
            } else {
              query = "(`" + _this11.type + "`.deleted = false OR `" + _this11.type + "`.deleted IS MISSING) ";
            }
          }
        }

        function formatSortComponent(sortComponent) {
          // Split the string by spaces to get individual parts
          var parts = sortComponent.split(" "); // Assuming the first part is the column name, add backticks around it

          parts[0] = "`" + parts[0] + "`"; // Rejoin the parts back into a string and return

          return parts.join(" ");
        }

        if ((0, _Security.hasSQLInjection)(query)) {
          return [];
        }

        if (args.limit) {
          args.limit = Number(args.limit);
        }

        if (args.offset) {
          args.offset = Number(args.offset);
        }

        if (!args.sort) {
          args.sort = "`" + _this11.type + "`.created_at DESC";
        } else {
          var sort_array = args.sort.split(","); // check the first string in each item in the array to see if it contains a . if not then add the type

          var new_sort_array = [];

          for (var i = 0; i < sort_array.length; i++) {
            if (!sort_array[i].includes(".")) {
              new_sort_array.push("`" + _this11.type + "`." + formatSortComponent(sort_array[i]));
            } else {
              new_sort_array.push(sort_array[i]);
            }
          }

          args.sort = _.join(new_sort_array, ",");
        }

        if (args.skip_hooks != true) {
          yield _this11.run_hook(_this11, "list", "before");
        }

        function removeSpaceAndSpecialCharacters(str) {
          return str.replace(/[^a-zA-Z0-9]/g, "");
        }

        var key = removeSpaceAndSpecialCharacters("list::" + _this11.type + "::" + args._join + "::" + query + "::" + args.limit + "::" + args.offset + "::" + args.sort + "::" + args.do_count + "::" + args.statement_consistent + "::" + args.columns + "::" + args.is_full_text + "::" + args.is_custom_query);
        var results;

        if (args.is_custom_query && args.is_custom_query === "true") {
          if (args.ids.length > 0) {
            if (_this11.shouldUseCache(_this11.type)) {
              var cached_results = yield _this11.getCacheProviderObject(_this11.type).get(key);
              results = cached_results == null ? void 0 : cached_results.value;

              if ((cached_results == null ? void 0 : cached_results.value) == undefined || (yield _this11.shouldForceRefresh(cached_results))) {
                results = yield _this11.database.query(query);

                _this11.getCacheProviderObject(_this11.type).set(key, {
                  value: results,
                  time: new Date().getTime()
                }, _this11.getCacheConfig(_this11.type));
              }
            } else {
              results = yield _this11.database.query(query);
            }
          }
        } else {
          if (args.is_full_text && args.is_full_text === "true") {
            if (_this11.shouldUseCache(_this11.type)) {
              var _cached_results = yield _this11.getCacheProviderObject(_this11.type).get(key);

              results = _cached_results == null ? void 0 : _cached_results.value;

              if ((_cached_results == null ? void 0 : _cached_results.value) == undefined || (yield _this11.shouldForceRefresh(_cached_results))) {
                results = yield _this11.database.full_text_search(_this11.type, query || "", args.limit, args.offset, args._join);

                _this11.getCacheProviderObject(_this11.type).set(key, {
                  value: results,
                  time: new Date().getTime()
                }, _this11.getCacheConfig(_this11.type));
              }
            } else {
              results = yield _this11.database.full_text_search(_this11.type, query || "", args.limit, args.offset, args._join);
            }
          } else {
            if (_this11.shouldUseCache(_this11.type)) {
              var _cached_results2 = yield _this11.getCacheProviderObject(_this11.type).get(key);

              results = _cached_results2 == null ? void 0 : _cached_results2.value;
              var shouleForceRefresh = yield _this11.shouldForceRefresh(_cached_results2);

              if ((_cached_results2 == null ? void 0 : _cached_results2.value) == undefined || shouleForceRefresh) {
                results = yield _this11.database.search(_this11.type, args.columns || "", query || "", args.limit, args.offset, args.sort, args.do_count, args.statement_consistent, args._join);

                _this11.getCacheProviderObject(_this11.type).set(key, {
                  value: results,
                  time: new Date().getTime()
                }, _this11.getCacheConfig(_this11.type));
              }
            } else {
              results = yield _this11.database.search(_this11.type, args.columns || "", query || "", args.limit, args.offset, args.sort, args.do_count, args.statement_consistent, args._join);
            }
          }
        }

        try {
          if (args.skip_read_serialize != true && args.skip_serialize != true) {
            results.data = yield _this11.do_serialize(results.data, "read", {}, args, (yield _this11.propsToBeRemoved(results.data)));
          }

          if (_this11.type == "resourcedetail") ;

          if (args.skip_hooks != true) {
            yield _this11.run_hook(results.data, "list", "after");
          }
        } catch (e) {
          console.log(e);
        }

        results.data = filterResultsByColumns(results.data, args.columns);
        return results;
      } catch (e) {
        console.log(e.stack);
        throw e;
      }
    })();
  }

  addHook(_ref2) {
    var {
      operation,
      when,
      execute
    } = _ref2;

    this[_hooks][operation][when].push(execute);
  }

  run_hook(data, op, when, old_data) {
    var _this12 = this;

    return _asyncToGenerator(function* () {
      try {
        if (_this12[_disable_lifecycle_events] == false) {
          var resourceLifecycleTriggered = new _ResourceLifecycleTriggered.default({
            data: {
              data,
              operation: op,
              when,
              old_data,
              resource: _this12.type,
              ctx: _this12[_ctx]
            }
          });
          resourceLifecycleTriggered.dispatch();
        }

        if (_this12[_hooks] && _this12[_hooks][op] && _this12[_hooks][op][when]) {
          for (var i of _this12[_hooks][op][when]) {
            data = yield i(data, old_data);
          }
        }

        return data;
      } catch (e) {
        throw e;
      }
    })();
  }

  shouldSerializerRun(args, type) {
    if (args) {
      if (args.skip_serialize == true) {
        return false;
      }

      if (type == "read") {
        if (args.skip_read_serialize == true) {
          return false;
        }
      } else if (type == "write") {
        if (args.skip_write_serialize == true) {
          return false;
        }
      }
    }

    return true;
  }

  mapToObject(data, Class, source_property, store_property, property) {
    var _this13 = this;

    return _asyncToGenerator(function* () {
      var original_is_array = _.isArray(data);

      if (!original_is_array) {
        data = Array.of(data);
      }

      var classes = _.isArray(Class) ? Class : [Class];
      var ids = [];

      _.each(data, result => {
        if (_.isString(result[source_property]) && result[source_property] != "") {
          ids = _.union(ids, [result[source_property]]);
        }
      });

      var returned_all = yield Promise.allSettled(_.map(classes, obj => {
        return new obj(_extends({}, _this13[_args], {
          skip_cache: _this13[_skip_cache]
        })).getMulti({
          skip_hooks: true,
          ids: ids
        });
      }));

      var ug = _.flatten(_.compact(_.map(returned_all, returned_obj => {
        if (returned_obj.status == "fulfilled") return returned_obj.value;
      })));

      data = _.map(data, result => {
        var result_found = _.find(ug, g => {
          return g.id == result[source_property];
        }) || {};
        result[store_property] = result_found;
        return result;
      });
      return original_is_array ? data : data[0];
    })();
  }

  mapToObjectArray(data, Class, source_property, store_property, property) {
    var _this14 = this;

    return _asyncToGenerator(function* () {
      var original_is_array = _.isArray(data);

      if (!original_is_array) {
        data = Array.of(data);
      }

      var ids = [];

      _.each(data, result => {
        var value = [];

        if (_.isArray(result[source_property])) {
          value = result[source_property];
        }

        if (_.isString(result[source_property])) {
          value = [result[source_property]];
        }

        var items = _.filter(value, obj => _.isString(obj) && obj != "");

        ids = _.union(ids, items);
      });

      var classes = _.isArray(Class) ? Class : [Class];
      var returned_all = yield Promise.allSettled(_.map(classes, obj => {
        return new obj(_extends({}, _this14[_args], {
          skip_cache: _this14[_skip_cache]
        })).getMulti({
          skip_hooks: true,
          ids: ids
        });
      }));

      var returned_objects = _.flatten(_.compact(_.map(returned_all, returned_obj => {
        if (returned_obj.status == "fulfilled") return returned_obj.value;
      })));
      /* let returned_objects = await new Class().getMulti({
        ids: ids,
      }); */


      _.each(data, result => {
        if (_.isString(result[store_property])) {
          result[store_property] = [result[store_property]];
        }

        if (!_.has(result, source_property)) {
          result[source_property] = [];
          return;
        }

        result[store_property] = _.map(result[source_property], pid => _.find(returned_objects, p => p.id === pid));
        result[store_property] = _.reject(result[store_property], obj => obj === null || obj === undefined);
      });

      return original_is_array ? data : data[0];
    })();
  }

  addModifier(_ref3) {
    var {
      when,
      execute
    } = _ref3;

    if (this[_serializers][when]) {
      this[_serializers][when]["modifiers"].push(execute);
    }
  }

  createModifiersFromProps() {
    this.createMofifier(this.props);
    this.addExternalModifiers();
  }

  addExternalModifiers(scope) {
    if (scope === void 0) {
      scope = "*";
    }

    var that = this;

    _.each([...spice.getModifiers(scope)], function (modifier) {
      that.addModifier(modifier);
    });
  }

  createMofifier(properties) {
    var _this15 = this;

    var _loop = function _loop(i) {
      if (properties[i].map) {
        switch (properties[i].map.type) {
          case _2.MapType.MODEL:
            {
              switch (properties[i].type) {
                case String:
                case "string":
                  {
                    _this15.addModifier({
                      when: properties[i].map.when || "read",
                      execute: function () {
                        var _execute = _asyncToGenerator(function* (data) {
                          return yield _this15.mapToObject(data, _.isString(properties[i].map.reference) ? spice.models[properties[i].map.reference] : properties[i].map.reference, i, properties[i].map.destination || i, properties[i]);
                        });

                        function execute(_x) {
                          return _execute.apply(this, arguments);
                        }

                        return execute;
                      }()
                    });

                    break;
                  }

                case Array:
                case "array":
                  {
                    _this15.addModifier({
                      when: properties[i].map.when || "read",
                      execute: function () {
                        var _execute2 = _asyncToGenerator(function* (data) {
                          return yield _this15.mapToObjectArray(data, _.isString(properties[i].map.reference) ? spice.models[properties[i].map.reference] : properties[i].map.reference, i, properties[i].map.destination || i, properties[i]);
                        });

                        function execute(_x2) {
                          return _execute2.apply(this, arguments);
                        }

                        return execute;
                      }()
                    });

                    break;
                  }
              }

              break;
            }

          case _2.MapType.LOOKUP:
            {
              break;
            }
        }
      }
    };

    for (var i in properties) {
      _loop(i);
    }
  }

  do_serialize(data, type, old_data, args, path_to_be_removed) {
    var _this16 = this;

    return _asyncToGenerator(function* () {
      //console.log("CTX INside Model DO", this[_ctx]);
      try {
        // run serializers
        if (!path_to_be_removed) {
          path_to_be_removed = [];
        }

        if (_this16.shouldSerializerRun(args, type)) {
          if (_this16.type != "" && _this16.type != undefined && _this16[_external_modifier_loaded] != true) {
            _this16.addExternalModifiers(_this16.type);

            _this16[_external_modifier_loaded] = true;
          }

          for (var i of _this16[_serializers][type]["modifiers"]) {
            try {
              data = yield i(data, old_data, _this16[_ctx], _this16.type);
            } catch (e) {
              console.log(e.stack);
            }
          } // make data an array


          var original_is_array = _.isArray(data);

          if (!original_is_array) {
            data = Array.of(data);
          } // get defaults from peroperties


          var defaults = {};

          for (var _i in _this16.props) {
            if (_this16.props[_i].defaults != undefined && _this16.props[_i].defaults[type] != undefined && _this16.props[_i].defaults[type] != undefined) {
              defaults[_i] = _.isFunction(_this16.props[_i].defaults[type]) ? _this16.props[_i].defaults[type]({
                old_data: data,
                new_data: old_data
              }) : _this16.props[_i].defaults[type];
            }
          } // apply defaults


          for (var _i2 of data) {
            _.defaults(_i2, defaults);
          } // apply cleaners


          if (type == "read") {
            var props_to_clean = ["deleted", "type", ...path_to_be_removed];

            for (var _i3 in _this16.props) {
              if (_this16.props[_i3].hide) {
                props_to_clean.push(_i3);
              }
            }

            data = _.map(data, obj => _.omit(obj, props_to_clean));
          }

          return original_is_array ? data : _.first(data);
        }
      } catch (e) {
        console.log(e.stack);
        throw e;
      }
    })();
  }

}

exports.default = SpiceModel;