"use strict";

exports.__esModule = true;
exports.default = void 0;

var _flat = _interopRequireDefault(require("flat"));

var _lodash = _interopRequireDefault(require("lodash"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function asyncGeneratorStep(gen, resolve, reject, _next, _throw, key, arg) { try { var info = gen[key](arg); var value = info.value; } catch (error) { reject(error); return; } if (info.done) { resolve(value); } else { Promise.resolve(value).then(_next, _throw); } }

function _asyncToGenerator(fn) { return function () { var self = this, args = arguments; return new Promise(function (resolve, reject) { var gen = fn.apply(self, args); function _next(value) { asyncGeneratorStep(gen, resolve, reject, _next, _throw, "next", value); } function _throw(err) { asyncGeneratorStep(gen, resolve, reject, _next, _throw, "throw", err); } _next(undefined); }); }; }

var fs = require("fs");

var path = require("path");

var {
  parse
} = require("json2csv");

class RestHelper {
  constructor() {}

  static clear_pagination(ctx, next) {
    return _asyncToGenerator(function* () {
      delete ctx.query.limit;
      delete ctx.query.offset;
      yield next();
    })();
  }

  static send_response(ctx, next) {
    return _asyncToGenerator(function* () {
      try {
        var response;

        if (_lodash.default.isArray(ctx.data)) {
          response = RestHelper.prepare_collection(RestHelper.SUCCESS, {
            data: ctx.data,
            total: ctx.total
          });
        } else {
          response = RestHelper.prepare_response(RestHelper.SUCCESS, ctx.data, ctx.errors, ctx.token);
        }

        ctx.status = 200;
        ctx.body = response;
      } catch (e) {
        console.log("Error at response: ", e, e.stack);
      }
    })();
  }

  static clear_pagination(ctx, next) {
    return _asyncToGenerator(function* () {
      delete ctx.query.limit;
      delete ctx.query.offset;
      yield next();
    })();
  }

  static send_download(ctx, next) {
    return _asyncToGenerator(function* () {
      function makeDirectory(dir) {
        if (!fs.existsSync(dir)) {
          fs.mkdirSync(dir);
        }
      }

      try {
        var download_type = ctx.request.query.format || "csv";
        var content;

        if (download_type == "csv") {
          var items = _lodash.default.map(ctx.data, item => {
            return (0, _flat.default)(item);
          });

          var fields = _lodash.default.union(_lodash.default.keys(_lodash.default.first(items)), _lodash.default.keys(_lodash.default.last(items)), _lodash.default.keys(_lodash.default.nth(items.length / 2)));

          var opts = {
            fields
          };
          content = parse(items, opts);
          ctx.set("content-type", "text/csv");
        } else {
          content = JSON.stringify(ctx.data);
          ctx.set("content-type", "application/json");
        }

        makeDirectory("./storage/exports");
        makeDirectory("./storage/exports/" + download_type + "/");
        var file = path.resolve("./storage/exports/" + download_type + "/" + RestHelper.makeid(9) + "." + download_type);
        yield fs.writeFile(file, content, function (err) {
          if (err) throw err;
        });
        ctx.set("content-disposition", "attachment");
        ctx.response.attachment(file);
        ctx.status = 200;
        ctx.body = fs.createReadStream(file);
      } catch (e) {
        console.log(e.stack);
        ctx.status = 400;
        ctx.body = RestHelper.prepare_response(RestHelper.FAILURE, e);
      }
    })();
  }

  static makeid(length) {
    var result = "";
    var characters = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789";
    var charactersLength = characters.length;

    for (var i = 0; i < length; i++) {
      result += characters.charAt(Math.floor(Math.random() * charactersLength));
    }

    return result;
  }

  static prepare_collection(status, obj) {
    if (obj === void 0) {
      obj = {
        data: [],
        total: null
      };
    }

    var data = {
      status: status,
      type: "collection",
      count: obj.data.length,
      data: obj.data,
      total: obj.total
    };

    if (obj.total == null || obj.total == undefined) {
      delete data.total;
    }

    return data;
  }

  static prepare_response(status, data, errors, token) {
    var response = {
      status: status,
      type: "single"
    };

    if (token) {
      response.token = token;
    }

    if (status == this.FAILURE) {
      response.message = data;

      if (errors) {
        response.errors = errors;
      }
    } else {
      if (data) {
        response.data = data;
      }
    }

    return response;
  }

  static get_typed_value(value) {
    var returnValue;

    try {
      returnValue = Number(value);

      if (isNaN(returnValue)) {
        throw returnValue;
      } else {
        return returnValue;
      }
    } catch (error) {
      try {
        if (value.toLowerCase() == "true" || value.toLowerCase() == "false") {
          returnValue = value === "true";
          return returnValue;
        } else {
          throw Error;
        }
      } catch (error) {
        try {
          returnValue = JSON.parse(value);
          return returnValue;
        } catch (error) {
          try {
            returnValue = String(value);
            return returnValue;
          } catch (error) {
            console.log("Cant parse" + value);
          }
        }
      }
    }
  }

  static normalizePort(val) {
    var port = parseInt(val, 10);

    if (isNaN(port)) {
      // named pipe
      return val;
    }

    if (port >= 0) {
      // port number
      return port;
    }

    return false;
  }

}

exports.default = RestHelper;
RestHelper.SUCCESS = "SUCCESS";
RestHelper.FAILURE = "FAILURE";