"use strict";

exports.__esModule = true;
exports.hasSQLInjection = hasSQLInjection;

function hasSQLInjection(input) {
  // Convert input to lowercase for easier pattern matching
  var lowerInput = input.toLowerCase(); // List of patterns that indicate potential SQL injection

  var patterns = ["--", // T-SQL comment
  "/*", // Start of block comment
  "*/", // End of block comment
  "or 1=1", // OR statement that's always true
  "or true", // OR statement that's always true (non-T-SQL databases)
  "union all", // UNION statement (often used to retrieve data from another table)
  "union select", // Another form of UNION statement
  ";", // Commenting out the rest of a query
  "drop table", // DDL statements are very suspicious
  "update ", // Modification commands
  "insert into", // Insertion commands
  "delete from", // Deletion commands
  "exec(", // Execution function in T-SQL
  "execute(", // Another execution function in T-SQL
  "xp_", // Extended stored procedures in T-SQL - often used in older SQL Server attacks
  "0x", // Hex encoding
  "benchmark(", // MySQL benchmark function - can be used for time-based blind attacks
  "sleep(", // MySQL sleep function - can also be used for time-based blind attacks
  "pg_sleep(", // PostgreSQL sleep function - time-based blind attack
  "cast(", // Casting can be used for type-based attacks
  "convert(", // Similar to the above
  "waitfor delay" // SQL Server delay function - time-based blind attack
  ];

  for (var pattern of patterns) {
    if (lowerInput.includes(pattern)) {
      return true;
    }
  } // Additional patterns based on regular expressions for more complex detections


  var regexPatterns = [/['"‘’“”][\s]*[+]+[\s]*['"‘’“”]/ // String concatenation, often used to obfuscate injections
  ];

  for (var regex of regexPatterns) {
    if (regex.test(lowerInput)) {
      return true;
    }
  }

  return false;
}