import path from "path";
import _, { indexOf } from "lodash";
import fs from "fs";
import uuid from "uuid";
//let paths = path.join(spice.root_path, "routes");
let tags = [];

const Comments = require("parse-comments");
const comments = new Comments();

function getPackageInfo() {
  let pac = require(path.join(spice.parent_app_dir, "/package.json"));
  return pac;
}

function createPath(path) {
  return _.join(
    _.map(_.split(path, "/"), (seg) => {
      if (_.startsWith(seg, ":")) {
        return `{${_.trim(seg, ":")}}`;
      }
      return seg;
    }),
    "/"
  );
}

function getRoutes(models) {
  let paths = path.join(spice.root_path, "routes");
  let routes = [];
  require("fs")
    .readdirSync(paths)
    .forEach(function (file) {
      if (file != "spice.js") {
        let router = require(path.join(paths, file));
        let router_string = fs.readFileSync(path.join(paths, file)).toString();
        let parsed_comments = comments.parse(router_string);

        let comment_processed = _.map(parsed_comments, (comment) => {
          if (comment.code.value.split(".")[0] == "router") {
            let code_array = comment.code.value.split(`"`);
            if (code_array.length == 1) {
              code_array = comment.code.value.split(`'`);
            }
            if (code_array[0] != undefined) {
              let router_method = code_array[0].split("(");
              let router_method_array = router_method[0].split(".");

              comment.path = createPath(code_array[1]);
              comment.original_path = code_array[1];
              comment.method = router_method_array[1];
              comment.path_method =
                comment.original_path + "-" + comment.method;
              return comment;
            }
          }
        }).filter((item) => item != undefined);
        tags.push({
          name: file.split(".")[0],
          description:
            parsed_comments[0] != undefined
              ? parsed_comments[0].description
              : "",
        });
        let defaultModel = findDefaultModel(parsed_comments[0]);
        _.each(router.stack, (layer) => {
          if (layer.methods.length > 0) {
            let path = {
              path: createPath(layer.path),
              original_path: layer.path,
              method: _.lowerCase(_.last(layer.methods)),
              params: _.map(layer.paramNames, (param) => {
                return {
                  name: param.name,
                  type: "string",
                  required: !param.optional,
                  in: "path",
                };
              }).filter((obj) => obj != undefined),
              tags: [file.split(".")[0]],
              produces: ["application/json"],
            };

            path.responses = {
              200: {
                description: "Successful Operation",
              },
              401: {
                description: "Authentication Error",
              },
              400: {
                description: "Bad Request",
              },
              404: {
                description: "Not found",
              },
            };

            let comment = _.find(comment_processed, (comment) => {
              function matches() {
                let match = false;
                match = _.endsWith(layer.path, comment.original_path);
                return match;
              }
              return (
                comment.method == _.lowerCase(_.last(layer.methods)) &&
                matches()
              );
            });

            if (comment != undefined) {
              path.description = comment.description;
              path.operationId = uuid.v4();
              path.parameters = [
                ...preparParameters(comment, undefined, models, router),
                ...path.params,
              ];
              let doc_return = filterTags(comment, "returns");
              if (doc_return.length > 0) {
                if (doc_return[0].type == undefined) {
                  path.responses["200"]["schema"] = {
                    $ref: `#/definitions/${doc_return[0].description}`,
                  };
                } else if (doc_return[0].type.name == "array") {
                  path.responses["200"]["schema"] = {
                    type: doc_return[0].type.name,
                    items: {
                      $ref: `#/definitions/${doc_return[0].description}`,
                    },
                  };
                } else if (doc_return[0].type.name == "object") {
                  path.responses["200"]["schema"] = {
                    type: doc_return[0].type.name,
                    additionalProperties: {
                      $ref: `#/definitions/${doc_return[0].description}`,
                    },
                  };
                }
              }
            }

            routes.push(path);
          }
        }).filter((item) => item != undefined);
      }
    });
  return routes;
}

function preparParameters(comment, defaultModel, models) {
  if (comment.method == "post" || comment.method == "put") {
    let param_list = filterTags(comment, "param");
    let route_model_params = [];
    let params_from_comments = _.map(param_list, (tag) => {
      let param;
      if (tag.name != "") {
        param = { name: tag.name, type: "string", paramType: "form" };
      }
      if (tag.description != undefined && IsJsonString(tag.description)) {
        let parsed_JSON = JSON.parse(tag.description);
        param = { ...param, ...parsed_JSON };
      }
      if (tag.name == "" && tag.type != undefined) {
        route_model_params.push(tag.type.name);
      }
      return param;
    }).filter((obj) => obj != undefined);
    let params_from_model = [];
    if (defaultModel && param_list.length == 0) {
      params_from_model = _.union(
        params_from_model,
        modelToParameter(defaultModel)
      );
    } else if (route_model_params.length > 0) {
      _.each(route_model_params, (Model) => {
        params_from_model = _.union(params_from_model, modelToParameter(Model));
      });
    }

    function modelToParameter(model) {
      return _.map(models[model], (value, key) => {
        return {
          name: key,
          type: _.includes(["object", "array"], value.type)
            ? "string"
            : value.type,
          description: value.description ? value.description : "",
          required: value.required ? value.required : false,
          paramType: "form",
          in: "formData",
        };
      }).filter((obj) => obj != undefined);
    }
    /* console.log("Props::::::: ", [
      ...params_from_comments,
      ...params_from_model,
    ]); */
    return [...params_from_comments, ...params_from_model];
  } else {
    return [];
  }
}

function IsJsonString(str) {
  try {
    JSON.parse(str);
  } catch (e) {
    return false;
  }
  return true;
}

function findDefaultModel(comment) {
  if (comment == undefined) {
    return undefined;
  } else {
    let id_found = _.first(filterTags(comment, "default_model"));
    return id_found == undefined ? id_found : id_found.description;
  }
}

function findId(comment) {
  let id_found = _.first(filterTags(comment, "id"));
  return id_found == undefined ? id_found : id_found.description;
}

function filterTags(comment, title) {
  return _.filter(findTags(comment), (tag) => {
    return tag.title == title;
  });
}

function findTags(comment) {
  return comment.tags;
}

async function getModels() {
  let paths = path.join(spice.root_path, "models");
  let models = {};
  require("fs")
    .readdirSync(paths)
    .forEach(function (file) {
      let model = require(path.join(paths, file));
      models[file.split(".")[0]] = model.props;
    });
  return models;
}

function getRequired(obj) {
  return _.map(obj, (value, key) => {
    if (value.required == true) {
      return key;
    }
  }).filter((item) => item != undefined);
}

function getProperties(obj) {
  return _.mapValues(obj, ({ type, description, options, schema, map }) => {
    let to_return = {
      type: type,
      description: description,
      enum: options,
    };

    if (to_return.type == "array") {
      to_return.items = {};
    }
    if (to_return.type == "date") {
      to_return.type = "string";
      to_return.format = "date";
    }

    if (map != undefined && map.type == "model") {
      if (map.reference != undefined) {
        let references = _.isArray(map.reference)
          ? map.reference
          : [map.reference];
        let reference = `#/definitions/${
          _.isString(map.reference) ? map.reference : references[0].name
        }`;
        if (to_return.type == "array") {
          to_return.items.$ref = reference;
        }

        if (to_return.type == "object") {
          to_return.$ref = reference;
        }

        if (to_return.type == "string") {
          to_return.type = "object";
          to_return.readOnly = true;
          to_return.$ref = reference;
        }
      }
    }
    return to_return;
  });
}

function generateDefinitions(items) {
  let models = _.cloneDeep(items);
  _.each(models, (value, key) => {
    _.each(value, (val, key2) => {
      if (val.props != undefined) {
        //console.log(key, key2, val, value);
        models[key2] = val.props;
        value[key2].map = {
          type: "model",
          reference: key2,
        };
      }
    });
  });
  let values = _.mapValues(models, (obj) => {
    return {
      type: "object",
      required: getRequired(obj),
      properties: getProperties(obj),
    };
  });

  return values;
}

function generatePaths(routes) {
  let json = _.groupBy(routes, "path");
  _.forEach(json, function (value, key) {
    json[key] = _.groupBy(json[key], "method");
    _.forEach(json[key], function (inside_value, inside_key) {
      json[key][inside_key] = inside_value[0];
    });
  });
  return json;
}

function generateTags() {
  return _.map(tags, (tag) => {
    return {
      name: tag.name,
      description: tag.description,
    };
  });
}

export default async function generateSpec() {
  let { description, version, name } = getPackageInfo();
  let models = _.cloneDeep(await getModels());
  let routes = getRoutes(models);

  let spec = {
    swagger: "2.0",
    info: {
      description: description,
      version: version,
      title: name,
      contact: {
        name: spice.config.spice.developer.name || "Sonover Inc.",
        email: spice.config.spice.developer.email,
      },
    },
    securityDefinitions: {
      Bearer: {
        type: "apiKey",
        name: "Authorization",
        in: "header",
      },
    },

    security: [
      {
        Bearer: [],
      },
    ],

    host: `${spice.config.spice.host}:${
      spice.config.spice.external_port || spice.config.spice.port
    }`,
    port: spice.config.spice.external_port || spice.config.spice.port,
    basePath: "/",
    tags: generateTags(),
    paths: generatePaths(routes),
    definitions: generateDefinitions(models),
  };
  return spec;
}
