"use strict";
var path = require("path");
//import IO from "koa-socket-2";
const cors = require("kcors");
global.spice = {};

require("dotenv").config();

var Validate = require("koa-validation");
const koaBody = require("koa-body");
spice.parent_app_dir = path.resolve("./");
spice.root_path = path.resolve("./build/");
spice.module_root_path = path.resolve(__dirname);
export { default as Event } from "./events/Event";
export { default as EventDebugger } from "./events/providers/EventDebugger";
export { default as SpiceEventDriver } from "./events/providers/SpiceEventDriver";
export { default as SocketEventDriver } from "./events/providers/SocketEventDriver";
export { default as ResourceLifecycleTriggered } from "./events/events/ResourceLifecycleTriggered";

export { default as SpiceModel } from "./models/SpiceModel";

export { default as Crypt } from "./utility/Crypt";

export { default as Serializer } from "./utility/Serializer";

export { default as RestHelper } from "./utility/RestHelper";
import Status from "./utility/Status";
export { default as Mail } from "./mail/Mail";
export { default as DataType } from "./utility/DataType";

export { default as MailDebug } from "./mail/providers/Debug";

export { default as DebugStorage } from "./storage/providers/Debug";
export { default as LocalStorage } from "./storage/providers/Local";
export { default as Storage } from "./storage/Storage";

export { default as MailFile } from "./mail/providers/File";
export { default as SpiceCache } from "./cache/providers/SpiceCache";
export { default as MapType } from "./utility/MapType";
export { default as addTask } from "./bootstrap/tasks";
const convert = require("koa-convert");
import { koaSwagger } from "koa2-swagger-ui";
spice.classes = {};
/* spice.classes.Event = Event;
spice.classes.Serializer = Serializer;
spice.classes.RestHelper = RestHelper; */
spice.classes.validate = Validate;
import docGen from "./docs/generator";
/* spice.classes.Mail = Mail; */
spice.classes.mail = {
  providers: {},
};
export let status = Status;
export default class Spice {
  constructor() {}

  async init(app) {
    try {
      //const io = new IO();
      app.use(cors());
      require("koa-qs")(app, "extended");

      app.use(
        koaBody({
          formLimit: "2gb",
          jsonLimit: "1gb",
          textLimit: "1gb",
          multipart: true,
          formidable: {
            uploadDir: "./storage/uploads",
            maxFileSize: 1024 * 1024 * 1024, // 1GB
          },
        })
      );
      app.use(convert(Validate()));
      spice.app = app;
      spice.mofifiers = {};
      spice.config = await require("./loaders/config")();

      spice.addModifier = function (resource, modifier) {
        spice.mofifiers[resource.toLowerCase()] =
          spice.mofifiers[resource] == undefined
            ? [modifier]
            : [...spice.mofifiers[resource], modifier];
      };

      spice.getModifiers = function (resource) {
        return spice.mofifiers[resource.toLowerCase()] || [];
      };
      //io.attach(app);
      //spice.io = io; */
      /* io.on("message", (ctx, data) => {
        console.log("client sent data to message endpoint", data);
      }); */
      /* app._io.on("connection", (sock) => {
        console.log("Connection Up", sock);
      }); */
      app.use(
        koaSwagger({
          hideTopbar: true,
          title: "Spice JS",
          routePrefix: "/docs", // host at /swagger instead of default /docs
          swaggerOptions: {
            spec: await docGen(),
            deepLinking: true,
            displayRequestDuration: true,
            jsonEditor: true,
          },
          exposeSpec: true,
        })
      );
      await require("./loaders").load();

      return spice;
    } catch (e) {
      console.log(e.stack);
    }
  }
}
