var views = require("koa-views");
const path = require("path");
const juice = require("juice");
import _ from "lodash";

export default class Mail {
  constructor(args = {}) {
    this.nunjucksOptions = {
      map: {
        html: "nunjucks",
      },
      options: {
        ext: spice.config.spice.view.ext || ".html",
        noCache: true,
        throwOnUndefined: false,
        globals: spice.config.spice.view.globals,
        filters: spice.config.spice.view.filters,
      },
    };

    if (spice.config.spice.view.custom_tags.enabled == true) {
      this.nunjucksOptions.options.tags = {
        variableStart: spice.config.spice.view.custom_tags.start || "${",
        variableEnd: spice.config.spice.view.custom_tags.end || "}",
      };
    }

    this.driver_name = args.driver_name || spice.config.mail.default_driver;
    this.driver_options = _.merge(
      spice.config.mail.drivers[this.driver_name],
      args.driver_options
    );
    this.template = args.template;
    this.Driver = spice.config.mail.providers[this.driver_name];
  }

  async prepare(data) {
    //apply template here
    let obj = { response: {} };
    views(path.resolve("./src/views"), this.nunjucksOptions)(obj, () => {});
    await obj.render(this.template, data);
    return obj.body;
  }

  async send({ data, options, template, content, driver_options }) {
    if (template) {
      this.template = template;
    }
    if (driver_options) {
      this.driver_options = _.merge(this.driver_options, driver_options);
    }
    let sendable_text = "";

    if (this.template) {
      sendable_text = await this.prepare({
        ...data,
        ...this,
      });
    }

    if (content) {
      sendable_text = content;
    }

    let driver = new this.Driver(this.driver_options);

    return await driver.send(juice(sendable_text), options);
  }
}
