"use strict";

import { MapType, DataType } from "..";
let co = require("co");
var regeneratorRuntime = require("regenerator-runtime");
import ResourceLifecycleTriggered from "../events/events/ResourceLifecycleTriggered";
import { hasSQLInjection } from "../utility/Security";

var SDate = require("sonover-date"),
  UUID = require("uuid"),
  _ = require("lodash"),
  _database = Symbol(),
  _ctx = Symbol(),
  _props = Symbol(),
  _args = Symbol(),
  _hooks = Symbol(),
  _disable_lifecycle_events = Symbol(),
  _external_modifier_loaded = Symbol(),
  _skip_cache = Symbol(),
  _serializers = Symbol();

//const _type = Symbol("type");
let that;
if (!Promise.allSettled) {
  Promise.allSettled = (promises) =>
    Promise.all(
      promises.map((promise, i) =>
        promise
          .then((value) => ({
            status: "fulfilled",
            value,
          }))
          .catch((reason) => ({
            status: "rejected",
            reason,
          }))
      )
    );
}
export default class SpiceModel {
  constructor(args = {}) {
    try {
      var dbtype =
        spice.config.database.connections[args.connection].type || "couchbase";
      let Database = require(`spice-${dbtype}`);
      this.type = "";
      this.collection = args.connection;
      this[_args] = args.args;
      this[_external_modifier_loaded] = false;
      this[_disable_lifecycle_events] =
        args.args?.disable_lifecycle_events || false;
      this[_ctx] = args?.args?.ctx;
      this[_skip_cache] = args?.args?.skip_cache || false;
      this[_hooks] = {
        create: {
          before: [],
          after: [],
        },
        get: {
          before: [],
          after: [],
        },
        list: {
          before: [],
          after: [],
        },
        update: {
          before: [],
          after: [],
        },
        delete: {
          before: [],
          after: [],
        },
      };
      this[_serializers] = {
        read: {
          modifiers: [],
          cleaners: ["deleted", "type"],
        },
        write: {
          modifiers: [],
          cleaners: [],
        },
      };

      this.deleted = false;
      this[_database] = new Database(args.connection || "default", {
        collection: args.collection,
        scope: args.scope,
      });

      function removeDynamicProps(props) {
        let returnVal = {};
        for (let i in props) {
          if (
            !_.has(props[i], "source") ||
            _.get(props[i], "source") == "static"
          ) {
            returnVal[i] = props[i];
          }
        }
        return returnVal;
      }

      function applySchemaOverrides(props, resource) {
        spice.schema_extenders = spice.schema_extenders || [];

        function applyExtenderModifier(prop, resource, extender) {
          if (extender.modifier) {
            return extender.modifier(prop, resource);
          }
          return prop;
        }

        function hasExtender(resource, prop, extender) {
          if (extender.selector(prop, resource)) {
            return true;
          }
          return false;
        }

        function resourceHasExtender(resource, extender) {
          return (
            extender.resource.includes(resource) ||
            extender.resource.includes("*")
          );
        }

        for (let file_name in spice.schema_extenders) {
          let extender = spice.schema_extenders[file_name];
          if (resourceHasExtender(resource, extender)) {
            for (let j in props) {
              if (hasExtender(resource, props[j], extender)) {
                props[j] = applyExtenderModifier(props[j], resource, extender);
              }
            }
            if (extender?.creator) {
              props = _.merge(props, extender.creator(resource));
            }
          }
        }
        return props;
      }
      //console.log("Applying Overrides to :", this.collection,'||', args.collection)
      args.props = applySchemaOverrides(
        removeDynamicProps(args.props),
        args.collection
      );
      this[_props] = args.props;
      for (let i in args.props) {
        if (args.args != undefined) {
          if (args.args[i] != undefined) {
            switch (args.props[i].type) {
              case String:
              case "string": {
                this[i] = args.args[i];
                break;
              }
              case Number:
              case "number": {
                this[i] = _.isNumber(args.args[i])
                  ? args.args[i]
                  : Number(args.args[i]);
                break;
              }
              case Boolean:
              case "boolean": {
                this[i] = _.isBoolean(args.args[i])
                  ? args.args[i]
                  : args.args[i] == "true" ||
                    args.args[i] == 1 ||
                    args.args[i] == "True";
                break;
              }
              case Date:
              case "date": {
                this[i] = args.args[i];
                break;
              }
              case Array:
              case "array": {
                this[i] = _.isArray(args.args[i])
                  ? args.args[i]
                  : JSON.parse(args.args[i]);
                break;
              }
              case Object:
              case "object": {
                this[i] = _.isObject(args.args[i])
                  ? args.args[i]
                  : JSON.parse(args.args[i]);
                break;
              }
              default: {
                this[i] = args.args[i];
              }
            }
          }
        }
      }
      that = this;
      this.createModifiersFromProps();
    } catch (e) {
      console.log(e.stack);
    }
    // }
  }

  get database() {
    return this[_database];
  }
  set database(value) {
    this[_database] = value;
  }

  get props() {
    return this[_props];
  }

  set props(value) {
    this[_props] = value;
  }

  is_date(item) {
    return item.length > 27 && item.indexOf("|") > -1;
  }

  async propsToBeRemoved(data) {
    if (!_.isArray(data)) {
      data = Array.of(data);
    }

    if (this[_ctx]) {
      if (this[_ctx]?.state?.process_fields) {
        let returned = await new this[_ctx].state.process_fields().process(
          this[_ctx],
          data,
          this.type
        );
        return returned;
      } else {
        return data;
      }
    }
    return [];
  }

  createValidationString({ property_name, props, operation, complete }) {
    let operationExempt = {
      PUT: ["required", "unique"],
    };
    let returnVal = [];
    for (let i in props) {
      if (!_.includes(operationExempt[operation] || [], i) || complete) {
        if (i == "required" && props[i] == true) {
          returnVal.push("required");
        }
        if (i == "email" && props[i] == true) {
          returnVal.push("email");
        }
        if (i == "json" && props[i] == true) {
          returnVal.push("json");
        }
        if (i == "ip" && props[i] == true) {
          returnVal.push("ip");
        }
        if (i == "url" && props[i] == true) {
          returnVal.push("url");
        }
        if (i == "date" && props[i] == true) {
          returnVal.push("date");
        }
        if (i == "requires") {
          returnVal.push(`shouldExist:${props[i]}`);
        }
        if (i == "unique" && props[i] == true) {
          returnVal.push(`unique:${this.constructor.name},${property_name}`);
        }
        if (i == "type") {
          switch (props[i]) {
            case String:
            case DataType.STRING: {
              returnVal.push(`string`);
              break;
            }
            case Number:
            case DataType.NUMBER: {
              //returnVal.push(`numeric`);
              break;
            }
            case Boolean:
            case DataType.BOOLEAN: {
              returnVal.push(`in:true,false,True,False,TRUE,FALSE,1,0`);
              break;
            }
            case Date:
            case DataType.DATE: {
              returnVal.push(`date`);
              break;
            }
            case Object:
            case DataType.OBJECT: {
              //returnVal.push(`json`);
              break;
            }
            case Array:
            case DataType.ARRAY: {
              //returnVal.push(`json`);
              break;
            }
            default: {
              returnVal.push(`string`);
              break;
            }
          }
        }
      }
    }
    return _.join(returnVal, "|");
  }

  createFilterString(property, props) {
    let returnVal = [];
    for (let i in props) {
      if (i == "lowercase" && props[i] == true) {
        returnVal.push("lowercase");
      }
      if (i == "trim" && props[i] == true) {
        returnVal.push("trim");
      }
      if (i == "uppercase" && props[i] == true) {
        returnVal.push("uppercase");
      }
    }
    return _.join(returnVal, "|");
  }

  getValidationRules(
    method = "POST",
    { complete = false, pick = [], omit = [] } = {}
  ) {
    let rules = {};
    let filters = {};
    let working_properties = this.props;
    if (pick.length > 0) {
      working_properties = _.pick(this.props, pick);
    }

    if (omit.length > 0) {
      working_properties = _.omit(this.props, omit);
    }

    for (let i in working_properties) {
      let created_rules = this.createValidationString({
        property_name: i,
        props: this.props[i],
        operation: method,
        complete,
      });
      if (created_rules.length > 0) {
        rules[i] = created_rules;
      }

      let created_filter = this.createFilterString(i, this.props[i]);
      if (created_filter.length > 0) {
        filters[i] = created_filter;
      }
    }
    return { rules, messages: {}, filters: { before: filters } };
  }

  makeQueryFromFilter(filter) {
    let return_string = "";
    for (let key in filter) {
      let item_string = "";
      if (filter[key].length > 1) {
        item_string = "(";
      }

      let item_count = 0;

      for (let item of filter[key]) {
        if (item_count > 0) {
          item_string = item_string + " OR ";
        }

        if (this.is_date(item)) {
          let to_index = item.indexOf("|");
          let first = item.substring(0, to_index);
          let second = item.substring(to_index + 1, item.length);
          item_string =
            item_string +
            " " +
            key +
            " BETWEEN '" +
            first +
            "' AND '" +
            second +
            "'";
        } else {
          item_string = item_string + " " + key + " LIKE '" + item + "'";
        }

        item_count++;
      }

      if (filter[key].length > 1) {
        item_string = item_string + ")";
      }
      if (return_string != "") {
        return_string = return_string + " AND ";
      }
      return_string = return_string + item_string;
    }
    return return_string;
  }

  shouldUseCache(resource_type) {
    // If '_skip_cache' property of this object is true, then we shouldn't cache.
    if (this[_skip_cache] == true) {
      return false;
    }

    // If the system configuration for spice has a cache status set to "disable",
    // then we shouldn't cache, so return false.
    if (spice.config.cache?.status == "disabled") {
      return false;
    }

    // If 'spice.cache[resource_type]' is not undefined,
    // it implies that this resource type is already in the cache or is cacheable.
    return spice.cache[resource_type] != undefined;
  }

  getCacheConfig(resource_type) {
    return spice.cache[resource_type] || {};
  }

  getCacheProviderObject(resource_type) {
    return spice.cache_providers[
      this.getCacheConfig(resource_type).driver ||
        this.getCacheConfig(resource_type).provider ||
        spice.config.cache.default_driver
    ];
  }

  async exists(item_type, key) {
    let obj = this.getCacheProviderObject(item_type);
    if (obj) return await obj.exists(key);
    return false;
  }

  async shouldForceRefresh(response) {
    let obj = this.getCacheProviderObject();
    let monitor_record = await obj.get(`monitor::${this.type}`);
    if (monitor_record == undefined) {
      return false;
    }
    if (monitor_record.time > response?.time) {
      return true;
    }
    return false;
  }

  setMonitor() {
    let current_time = new Date().getTime();
    let obj = this.getCacheProviderObject();
    let key = `monitor::${this.type}`;
    let value = { time: current_time };
    obj.set(key, value, { ttl: 0 });
  }

  async get(args) {
    try {
      if (!args) {
        args = {};
      }
      if (_.isString(args.id)) {
        if (args.skip_hooks != true) {
          await this.run_hook(args, "get", "before");
        }
        let key = `get::${this.type}::${args.id}`;
        let results = {};
        if (this.shouldUseCache(this.type)) {
          let cached_results = await this.getCacheProviderObject(this.type).get(
            key
          );

          results = cached_results?.value;
          if (
            cached_results?.value == undefined ||
            (await this.shouldForceRefresh(cached_results)) ||
            this.type == "workflow"
          ) {
            results = await this.database.get(args.id);
            await this.getCacheProviderObject(this.type).set(
              key,
              { value: results, time: new Date().getTime() },
              this.getCacheConfig(this.type)
            );
          }
        } else {
          results = await this.database.get(args.id);
        }

        if (results.type != undefined) {
          if (results.type != this.type)
            throw new Error(`${this.type} does not exist type`);
        }

        if (results._type != undefined) {
          if (results._type != this.type)
            throw new Error(`${this.type} does not exist _type`);
        }

        if (results.deleted == undefined || results.deleted == false) {
          if (args.skip_read_serialize != true && args.skip_serialize != true) {
            results = await this.do_serialize(
              results,
              "read",
              {},
              args,
              await this.propsToBeRemoved(results)
            );
          }
          if (args.skip_hooks != true) {
            await this.run_hook(results, "get", "after");
          }
          return results;
        } else {
          throw new Error(`${this.type} does not exist`);
        }
      }
    } catch (e) {
      console.log(e.message, e);
      throw e;
    }
  }

  async query(query, scope) {
    try {
      let results = await this.database.query(query, scope);
      return results;
    } catch (error) {
      throw error;
    }
  }

  async getMulti(args) {
    try {
      if (!args) {
        args = {};
      }
      if (args.skip_hooks != true) {
        await this.run_hook(this, "list", "before");
      }
      _.remove(args.ids, (o) => o == undefined);
      let key = `multi-get::${this.type}::${_.join(args.ids, "|")}`;
      let results = [];
      if (args.ids.length > 0) {
        if (this.shouldUseCache(this.type)) {
          let cached_results = await this.getCacheProviderObject(this.type).get(
            key
          );
          results = cached_results?.value;
          if (
            cached_results?.value == undefined ||
            (await this.shouldForceRefresh(cached_results))
          ) {
            results = await this.database.multi_get(args.ids, true);
            this.getCacheProviderObject(this.type).set(
              key,
              { value: results, time: new Date().getTime() },
              this.getCacheConfig(this.type)
            );
          }
        } else {
          results = await this.database.multi_get(args.ids, true);
        }
      }
      _.remove(results, (o) => o.type != this.type);
      if (args.skip_read_serialize != true && args.skip_serialize != true) {
        results = await this.do_serialize(
          results,
          "read",
          {},
          args,
          await this.propsToBeRemoved(results)
        );
      }

      if (args.skip_hooks != true) {
        await this.run_hook(results, "list", "after", args.context);
      }

      return results;
    } catch (e) {
      console.log(e.stack);
      throw e;
    }
  }

  async exist(data) {
    try {
      if (_.isString(data)) {
        let result = await this.database.get(data);
        if (result.type)
          if (result.type != this.type) {
            return false;
          }
        if (result._type)
          if (result._type != this.type) {
            return false;
          }
      } else {
        if (data.type)
          if (data.type != this.type) {
            return false;
          }

        if (data._type)
          if (data._type != this.type) {
            return false;
          }
      }
      return true;
    } catch (e) {
      throw e;
    }
  }

  async update(args) {
    try {
      this.updated_at = new SDate().now();
      let results = await this.database.get(args.id);
      let item_exist = await this.exist(results);
      if (!item_exist) {
        throw new Error(`${this.type} does not exist. in update`);
      }
      delete results["id"];
      _.defaults(this, results);
      let cover_obj = {
        old: results,
        new: this,
        id: args.id,
      };

      let form;
      if (args.skip_write_serialize != true && args.skip_serialize != true) {
        cover_obj.new = await this.do_serialize(
          this,
          "write",
          cover_obj.new,
          args
        );
      }

      if (args.skip_hooks != true) {
        await this.run_hook(cover_obj, "update", "before", results);
      }
      let db_data = cover_obj.new || this;

      await this.database.update(args.id, db_data);
      this.setMonitor();

      if (args.skip_read_serialize != true && args.skip_serialize != true) {
        cover_obj.new = await this.do_serialize(
          cover_obj.new,
          "read",
          {},
          args,
          await this.propsToBeRemoved(cover_obj.new)
        );
      }

      if (args.skip_hooks != true) {
        await this.run_hook(
          {
            ...this,
            id: args.id,
          },
          "update",
          "after",
          results
        );
      }
      this.id = args.id;
      return { ...cover_obj.new, id: args.id };
    } catch (e) {
      console.log("Error on update", e, e.stack);
      throw e;
    }
  }

  async create(args = {}) {
    try {
      let form;
      this.created_at = new SDate().now();
      args = _.defaults(args, { id_prefix: this.type });
      if (args.body) {
        form = _.defaults({}, args.body);
        form.created_at = this.created_at;
        form.updated_at = this.created_at;
        delete form["bucket"];
      }
      let workingForm = form || this;
      this.updated_at = new SDate().now();

      let id = `${args.id_prefix}-${UUID.v4()}`;
      if (args && args.id) {
        id = args.id;
      }

      if (args.skip_write_serialize != true && args.skip_serialize != true) {
        workingForm = await this.do_serialize(workingForm, "write", {}, args);
      }

      if (args.skip_hooks != true) {
        await this.run_hook(workingForm, "create", "before");
      }

      let results = await this.database.insert(id, workingForm, args.expiry);
      this.setMonitor();

      if (args.skip_read_serialize != true && args.skip_serialize != true) {
        results = await this.do_serialize(
          results,
          "read",
          {},
          args,
          await this.propsToBeRemoved(results)
        );
      }
      if (args.skip_hooks != true) {
        await this.run_hook(
          {
            ...results,
            id,
          },
          "create",
          "after"
        );
      }
      return { ...results, id };
    } catch (e) {
      console.log(e.stack);
      throw e;
    }
  }

  async delete(args) {
    let item_exist = await this.exist(args.id);

    if (!item_exist) {
      throw new Error(`${this.type} does not exist.`);
    }
    let results = await this.database.get(args.id);
    try {
      if (args.skip_hooks != true) {
        await this.run_hook(args, "delete", "before");
      }
      let delete_response = {};
      if (args.hard) {
        delete_response = await this.database.delete(args.id);
        this.setMonitor();
      } else {
        delete results["id"];
        results.deleted = true;
        delete_response = await this.database.update(args.id, "");
      }
      if (args.skip_hooks != true) {
        await this.run_hook(results, "delete", "after", results);
      }
      return {};
    } catch (e) {
      console.log(e.stack);
      throw e;
    }
  }

  toString() {
    JSON.stringify(this);
  }

  async list(args) {
    try {
      if (!args) {
        args = {};
      }

      function prepColumns(columns) {
        if (columns && columns != "") {
          let columnList = columns.split(","); // Convert comma-separated string to array
          return _.join(
            _.compact(
              _.map(columnList, (column) => {
                if (column == "meta().id") {
                  column = undefined;
                } else if (!column.startsWith("`") && !column.endsWith("`")) {
                  `\`${column}\``;
                }
                if (column) {
                  if (!column.includes(".") && !column.startsWith(that.type)) {
                    column = `\`${that.type}\`.${column}`;
                  }
                }
                return column;
              })
            ),
            ","
          );
        }
        return columns;
      }

      let query = "";
      args.columns = prepColumns(args.columns);
      function filterResultsByColumns(data, columns) {
        if (columns && columns != "") {
          columns = columns.replace(/`/g, "");
          columns = columns.replace(/meta\(\)\.id/g, "id");
          let columnList = columns.split(",").map((column) => {
            //if colums has a . then remove the first part of the string
            return column.includes(".") ? _.last(column.split(".")) : column;
          }); // Convert comma-separated string to array
          columnList.push("_permissions"); // Always include _permissions
          columnList.push("_permissions_"); // Always include _permissions
          columnList.push("id"); // Always include _permissions
          return _.map(data, (entry) => _.pick(entry, columnList));
        }
        return data;
      }
      function extractNestings(string, localType) {
        let returnVal = [];
        let regex = /(\w+)\./g;
        let match;
        while ((match = regex.exec(string))) {
          if (match[1] != localType) returnVal.push(match[1]);
        }
        return returnVal;
      }

      let nestings = [
        ...extractNestings(args?.query, this.type),
        ...extractNestings(args?.columns, this.type),
        ...extractNestings(args?.sort, this.type),
      ];
      let mappedNestings = _.compact(
        _.map(_.uniq(nestings), (nesting) => {
          let prop = this.props[nesting];
          if (prop) {
            if (prop.map) {
              if (prop.map.type == MapType.MODEL) {
                return {
                  alias: nesting,
                  reference: prop.map.reference.toLowerCase(),
                  value_field: prop.map.value_field,
                };
              }
            }
          }
        })
      );

      function createJoinSection(nestings) {
        //create join section with nestings
        let joinSection = "";
        _.each(nestings, (nesting) => {
          joinSection += `LEFT JOIN \`${nesting.reference}\` AS \`${nesting.alias}\` ON KEYS \`${that.type}\`.\`${nesting.alias}\` `;
        });
        return joinSection;
      }
      let _join = createJoinSection(mappedNestings);
      args._join = _join;

      if (
        (args.is_full_text && args.is_full_text === "true") ||
        (args.is_custom_query && args.is_custom_query === "true")
      ) {
        query = args.query;
      } else {
        if (args.filters) {
          query = this.makeQueryFromFilter(args.filters);
        } else {
          if (args.query) {
            query =
              args.query +
              ` AND (\`${this.type}\`.deleted = false OR \`${this.type}\`.deleted IS MISSING) `;
          } else {
            query = `(\`${this.type}\`.deleted = false OR \`${this.type}\`.deleted IS MISSING) `;
          }
        }
      }

      function formatSortComponent(sortComponent) {
        // Split the string by spaces to get individual parts
        const parts = sortComponent.split(" ");
        // Assuming the first part is the column name, add backticks around it
        parts[0] = `\`${parts[0]}\``;
        // Rejoin the parts back into a string and return
        return parts.join(" ");
      }

      if (hasSQLInjection(query)) {
        return [];
      }
      if (args.limit) {
        args.limit = Number(args.limit);
      }
      if (args.offset) {
        args.offset = Number(args.offset);
      }

      if (!args.sort) {
        args.sort = `\`${this.type}\`.created_at DESC`;
      } else {
        let sort_array = args.sort.split(",");
        // check the first string in each item in the array to see if it contains a . if not then add the type
        let new_sort_array = [];
        for (let i = 0; i < sort_array.length; i++) {
          if (!sort_array[i].includes(".")) {
            new_sort_array.push(
              `\`${this.type}\`.${formatSortComponent(sort_array[i])}`
            );
          } else {
            new_sort_array.push(sort_array[i]);
          }
        }
        args.sort = _.join(new_sort_array, ",");
      }
      if (args.skip_hooks != true) {
        await this.run_hook(this, "list", "before");
      }
      function removeSpaceAndSpecialCharacters(str) {
        return str.replace(/[^a-zA-Z0-9]/g, "");
      }
      let key = removeSpaceAndSpecialCharacters(
        `list::${this.type}::${args._join}::${query}::${args.limit}::${args.offset}::${args.sort}::${args.do_count}::${args.statement_consistent}::${args.columns}::${args.is_full_text}::${args.is_custom_query}`
      );

      let results;
      if (args.is_custom_query && args.is_custom_query === "true") {
        if (args.ids.length > 0) {
          if (this.shouldUseCache(this.type)) {
            let cached_results = await this.getCacheProviderObject(
              this.type
            ).get(key);
            results = cached_results?.value;
            if (
              cached_results?.value == undefined ||
              (await this.shouldForceRefresh(cached_results))
            ) {
              results = await this.database.query(query);
              this.getCacheProviderObject(this.type).set(
                key,
                { value: results, time: new Date().getTime() },
                this.getCacheConfig(this.type)
              );
            }
          } else {
            results = await this.database.query(query);
          }
        }
      } else {
        if (args.is_full_text && args.is_full_text === "true") {
          if (this.shouldUseCache(this.type)) {
            let cached_results = await this.getCacheProviderObject(
              this.type
            ).get(key);
            results = cached_results?.value;
            if (
              cached_results?.value == undefined ||
              (await this.shouldForceRefresh(cached_results))
            ) {
              results = await this.database.full_text_search(
                this.type,
                query || "",
                args.limit,
                args.offset,
                args._join
              );
              this.getCacheProviderObject(this.type).set(
                key,
                { value: results, time: new Date().getTime() },
                this.getCacheConfig(this.type)
              );
            }
          } else {
            results = await this.database.full_text_search(
              this.type,
              query || "",
              args.limit,
              args.offset,
              args._join
            );
          }
        } else {
          if (this.shouldUseCache(this.type)) {
            let cached_results = await this.getCacheProviderObject(
              this.type
            ).get(key);
            results = cached_results?.value;
            let shouleForceRefresh =
              await this.shouldForceRefresh(cached_results);
            if (cached_results?.value == undefined || shouleForceRefresh) {
              results = await this.database.search(
                this.type,
                args.columns || "",
                query || "",
                args.limit,
                args.offset,
                args.sort,
                args.do_count,
                args.statement_consistent,
                args._join
              );
              this.getCacheProviderObject(this.type).set(
                key,
                { value: results, time: new Date().getTime() },
                this.getCacheConfig(this.type)
              );
            }
          } else {
            results = await this.database.search(
              this.type,
              args.columns || "",
              query || "",
              args.limit,
              args.offset,
              args.sort,
              args.do_count,
              args.statement_consistent,
              args._join
            );
          }
        }
      }
      try {
        if (args.skip_read_serialize != true && args.skip_serialize != true) {
          results.data = await this.do_serialize(
            results.data,
            "read",
            {},
            args,
            await this.propsToBeRemoved(results.data)
          );
        }
        if (this.type == "resourcedetail");

        if (args.skip_hooks != true) {
          await this.run_hook(results.data, "list", "after");
        }
      } catch (e) {
        console.log(e);
      }
      results.data = filterResultsByColumns(results.data, args.columns);
      return results;
    } catch (e) {
      console.log(e.stack);
      throw e;
    }
  }

  addHook({ operation, when, execute }) {
    this[_hooks][operation][when].push(execute);
  }

  async run_hook(data, op, when, old_data) {
    try {
      if (this[_disable_lifecycle_events] == false) {
        let resourceLifecycleTriggered = new ResourceLifecycleTriggered({
          data: {
            data,
            operation: op,
            when,
            old_data,
            resource: this.type,
            ctx: this[_ctx],
          },
        });
        resourceLifecycleTriggered.dispatch();
      }
      if (this[_hooks] && this[_hooks][op] && this[_hooks][op][when]) {
        for (let i of this[_hooks][op][when]) {
          data = await i(data, old_data);
        }
      }
      return data;
    } catch (e) {
      throw e;
    }
  }

  shouldSerializerRun(args, type) {
    if (args) {
      if (args.skip_serialize == true) {
        return false;
      }
      if (type == "read") {
        if (args.skip_read_serialize == true) {
          return false;
        }
      } else if (type == "write") {
        if (args.skip_write_serialize == true) {
          return false;
        }
      }
    }
    return true;
  }

  async mapToObject(data, Class, source_property, store_property, property) {
    let original_is_array = _.isArray(data);
    if (!original_is_array) {
      data = Array.of(data);
    }

    let classes = _.isArray(Class) ? Class : [Class];

    let ids = [];
    _.each(data, (result) => {
      if (
        _.isString(result[source_property]) &&
        result[source_property] != ""
      ) {
        ids = _.union(ids, [result[source_property]]);
      }
    });

    var returned_all = await Promise.allSettled(
      _.map(classes, (obj) => {
        return new obj({
          ...this[_args],
          skip_cache: this[_skip_cache],
        }).getMulti({
          skip_hooks: true,
          ids: ids,
        });
      })
    );
    let ug = _.flatten(
      _.compact(
        _.map(returned_all, (returned_obj) => {
          if (returned_obj.status == "fulfilled") return returned_obj.value;
        })
      )
    );

    data = _.map(data, (result) => {
      let result_found =
        _.find(ug, (g) => {
          return g.id == result[source_property];
        }) || {};
      result[store_property] = result_found;
      return result;
    });
    return original_is_array ? data : data[0];
  }

  async mapToObjectArray(
    data,
    Class,
    source_property,
    store_property,
    property
  ) {
    let original_is_array = _.isArray(data);
    if (!original_is_array) {
      data = Array.of(data);
    }
    let ids = [];
    _.each(data, (result) => {
      let value = [];

      if (_.isArray(result[source_property])) {
        value = result[source_property];
      }

      if (_.isString(result[source_property])) {
        value = [result[source_property]];
      }

      let items = _.filter(value, (obj) => _.isString(obj) && obj != "");
      ids = _.union(ids, items);
    });

    let classes = _.isArray(Class) ? Class : [Class];
    var returned_all = await Promise.allSettled(
      _.map(classes, (obj) => {
        return new obj({
          ...this[_args],
          skip_cache: this[_skip_cache],
        }).getMulti({
          skip_hooks: true,
          ids: ids,
        });
      })
    );

    var returned_objects = _.flatten(
      _.compact(
        _.map(returned_all, (returned_obj) => {
          if (returned_obj.status == "fulfilled") return returned_obj.value;
        })
      )
    );

    /* let returned_objects = await new Class().getMulti({
      ids: ids,
    }); */
    _.each(data, (result) => {
      if (_.isString(result[store_property])) {
        result[store_property] = [result[store_property]];
      }

      if (!_.has(result, source_property)) {
        result[source_property] = [];
        return;
      }

      result[store_property] = _.map(result[source_property], (pid) =>
        _.find(returned_objects, (p) => p.id === pid)
      );
      result[store_property] = _.reject(
        result[store_property],
        (obj) => obj === null || obj === undefined
      );
    });
    return original_is_array ? data : data[0];
  }

  addModifier({ when, execute }) {
    if (this[_serializers][when]) {
      this[_serializers][when]["modifiers"].push(execute);
    }
  }

  createModifiersFromProps() {
    this.createMofifier(this.props);
    this.addExternalModifiers();
  }

  addExternalModifiers(scope = "*") {
    let that = this;
    _.each([...spice.getModifiers(scope)], function (modifier) {
      that.addModifier(modifier);
    });
  }

  createMofifier(properties) {
    for (let i in properties) {
      if (properties[i].map) {
        switch (properties[i].map.type) {
          case MapType.MODEL: {
            switch (properties[i].type) {
              case String:
              case "string": {
                this.addModifier({
                  when: properties[i].map.when || "read",
                  execute: async (data) => {
                    return await this.mapToObject(
                      data,
                      _.isString(properties[i].map.reference)
                        ? spice.models[properties[i].map.reference]
                        : properties[i].map.reference,
                      i,
                      properties[i].map.destination || i,
                      properties[i]
                    );
                  },
                });
                break;
              }
              case Array:
              case "array": {
                this.addModifier({
                  when: properties[i].map.when || "read",
                  execute: async (data) => {
                    return await this.mapToObjectArray(
                      data,
                      _.isString(properties[i].map.reference)
                        ? spice.models[properties[i].map.reference]
                        : properties[i].map.reference,
                      i,
                      properties[i].map.destination || i,
                      properties[i]
                    );
                  },
                });
                break;
              }
            }
            break;
          }
          case MapType.LOOKUP: {
            break;
          }
        }
      }
    }
  }

  async do_serialize(data, type, old_data, args, path_to_be_removed) {
    //console.log("CTX INside Model DO", this[_ctx]);
    try {
      // run serializers

      if (!path_to_be_removed) {
        path_to_be_removed = [];
      }

      if (this.shouldSerializerRun(args, type)) {
        if (
          this.type != "" &&
          this.type != undefined &&
          this[_external_modifier_loaded] != true
        ) {
          this.addExternalModifiers(this.type);
          this[_external_modifier_loaded] = true;
        }
        for (let i of this[_serializers][type]["modifiers"]) {
          try {
            data = await i(data, old_data, this[_ctx], this.type);
          } catch (e) {
            console.log(e.stack);
          }
        }

        // make data an array
        let original_is_array = _.isArray(data);

        if (!original_is_array) {
          data = Array.of(data);
        }

        // get defaults from peroperties
        let defaults = {};
        for (let i in this.props) {
          if (
            this.props[i].defaults != undefined &&
            this.props[i].defaults[type] != undefined &&
            this.props[i].defaults[type] != undefined
          ) {
            defaults[i] = _.isFunction(this.props[i].defaults[type])
              ? this.props[i].defaults[type]({
                  old_data: data,
                  new_data: old_data,
                })
              : this.props[i].defaults[type];
          }
        }
        // apply defaults
        for (let i of data) {
          _.defaults(i, defaults);
        }

        // apply cleaners
        if (type == "read") {
          let props_to_clean = ["deleted", "type", ...path_to_be_removed];
          for (let i in this.props) {
            if (this.props[i].hide) {
              props_to_clean.push(i);
            }
          }
          data = _.map(data, (obj) => _.omit(obj, props_to_clean));
        }

        return original_is_array ? data : _.first(data);
      }
    } catch (e) {
      console.log(e.stack);
      throw e;
    }
  }
}
