import flatten from "flat";
import _ from "lodash";
const fs = require("fs");
const path = require("path");

const { parse } = require("json2csv");
export default class RestHelper {
    constructor() { }
    static async clear_pagination(ctx, next) {
        delete ctx.query.limit;
        delete ctx.query.offset;
        await next();
    }

    static async send_response(ctx, next) {
        try {
            let response;
            if (_.isArray(ctx.data)) {
                response = RestHelper.prepare_collection(RestHelper.SUCCESS, {
                    data: ctx.data,
                    total: ctx.total
                });
            } else {
                response = RestHelper.prepare_response(RestHelper.SUCCESS, ctx.data, ctx.errors, ctx.token);
            }
            ctx.status = 200;
            ctx.body = response;
        } catch (e) {
            console.log("Error at response: ", e, e.stack);
        }
    }
  

    static async clear_pagination(ctx, next) {
        delete ctx.query.limit;
        delete ctx.query.offset;
        await next();
    }

  static async send_download(ctx, next) {
    function makeDirectory(dir) {
      if (!fs.existsSync(dir)) {
        fs.mkdirSync(dir);
      }
    }
    try {
      let download_type = ctx.request.query.format || "csv";
      let content;
      if (download_type == "csv") {
        let items = _.map(ctx.data, (item) => {
          return flatten(item);
        });
        let fields = _.union(
          _.keys(_.first(items)),
          _.keys(_.last(items)),
          _.keys(_.nth(items.length / 2))
        );
        const opts = {
          fields,
        };
        content = parse(items, opts);
        ctx.set("content-type", "text/csv");
      } else {
        content = JSON.stringify(ctx.data);
        ctx.set("content-type", "application/json");
      }

      makeDirectory(`./storage/exports`);
      makeDirectory(`./storage/exports/${download_type}/`);
      let file = path.resolve(
        `./storage/exports/${download_type}/${RestHelper.makeid(
          9
        )}.${download_type}`
      );
      await fs.writeFile(file, content, function (err) {
        if (err) throw err;
      });

      ctx.set("content-disposition", "attachment");
      ctx.response.attachment(file);
      ctx.status = 200;
      ctx.body = fs.createReadStream(file);
    } catch (e) {
      console.log(e.stack);
      ctx.status = 400;
      ctx.body = RestHelper.prepare_response(RestHelper.FAILURE, e);
    }
  }

  static makeid(length) {
    var result = "";
    var characters =
      "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789";
    var charactersLength = characters.length;
    for (var i = 0; i < length; i++) {
      result += characters.charAt(Math.floor(Math.random() * charactersLength));
    }
    return result;
  }

  static SUCCESS = "SUCCESS";
  static FAILURE = "FAILURE";
  static prepare_collection(
    status,
    obj = {
      data: [],
      total: null,
    }
  ) {
    var data = {
      status: status,
      type: "collection",
      count: obj.data.length,
      data: obj.data,
      total: obj.total,
    };
    if (obj.total == null || obj.total == undefined) {
      delete data.total;
    }
    return data;
  }

  static prepare_response(status, data, errors, token) {
    var response = {
      status: status,
      type: "single",
    };

    if (token) {
      response.token = token;
    }

    if (status == this.FAILURE) {
      response.message = data;
      if (errors) {
        response.errors = errors;
      }
    } else {
      if (data) {
        response.data = data;
      }
    }

    return response;
  }

  static get_typed_value(value) {
    var returnValue;
    try {
      returnValue = Number(value);
      if (isNaN(returnValue)) {
        throw returnValue;
      } else {
        return returnValue;
      }
    } catch (error) {
      try {
        if (value.toLowerCase() == "true" || value.toLowerCase() == "false") {
          returnValue = value === "true";
          return returnValue;
        } else {
          throw Error;
        }
      } catch (error) {
        try {
          returnValue = JSON.parse(value);
          return returnValue;
        } catch (error) {
          try {
            returnValue = String(value);
            return returnValue;
          } catch (error) {
            console.log("Cant parse" + value);
          }
        }
      }
    }
  }

  static normalizePort(val) {
    var port = parseInt(val, 10);

    if (isNaN(port)) {
      // named pipe
      return val;
    }

    if (port >= 0) {
      // port number
      return port;
    }

    return false;
  }
}
